#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import copy
import numbers

import numpy

import histbook.axis

class Channel(object):
    """Abstract class for graphical channels in a Vega-Lite plot."""

    def __init__(self, axis):
        self.axis = axis

class OverlayChannel(Channel):
    """Represents an overlayed axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".overlay({0})".format(self.axis)

class StackChannel(Channel):
    """Represents a stacked axis in a Vega-Lite plot."""

    def __init__(self, axis, order):
        super(StackChannel, self).__init__(axis)
        self.order = order

    def __repr__(self):
        if self.order is None:
            return ".stack({0})".format(self.axis)
        else:
            return ".stack({0}, order={1})".format(self.axis, repr(self.order))

class BesideChannel(Channel):
    """Represents a side-by-side axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".beside({0})".format(self.axis)

class BelowChannel(Channel):
    """Represents an above-and-below axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".below({0})".format(self.axis)

class Terminal1dChannel(Channel):
    """Abstract class for the last graphical channel in a 1d Vega-Lite plot."""

    def __init__(self, axis, profile, error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale):
        self.axis = axis
        self.profile = profile
        self.error = error
        self.normalized = normalized
        self.width = width
        self.height = height
        self.title = title
        self.config = config
        self.xscale = xscale
        self.yscale = yscale
        self.colorscale = colorscale
        self.shapescale = shapescale

    def __repr__(self):
        args = [repr(self.axis)]
        if self.profile is not None:
            args.append("profile={0}".format(self.profile))
        if (not isinstance(self, MarkerChannel) and self.error is not False) or (isinstance(self, MarkerChannel) and self.error is not True):
            args.append("error={0}".format(self.error))
        if self.normalized is not False:
            args.append("normalized={0}".format(self.normalized))
        if self.width is not None:
            args.append("width={0}".format(repr(self.width)))
        if self.height is not None:
            args.append("height={0}".format(repr(self.height)))
        if self.title is not None:
            args.append("title={0}".format(repr(self.title)))
        if self.config is not None:
            args.append("config={0}".format(repr(self.config)))
        if self.xscale is not None:
            args.append("xscale={0}".format(repr(self.xscale)))
        if self.yscale is not None:
            args.append("yscale={0}".format(repr(self.yscale)))
        if self.colorscale is not None:
            args.append("colorscale={0}".format(repr(self.colorscale)))
        if self.shapescale is not None:
            args.append("shapescale={0}".format(repr(self.shapescale)))
        return ".{0}({1})".format(self._method, "".join(args))

class BarChannel(Terminal1dChannel):
    """Represents a bar axis in a Vega-Lite plot."""
    _method = "bar"

class StepChannel(Terminal1dChannel):
    """Represents a step axis in a Vega-Lite plot."""
    _method = "step"

class AreaChannel(Terminal1dChannel):
    """Represents an area axis in a Vega-Lite plot."""
    _method = "area"

class LineChannel(Terminal1dChannel):
    """Represents a line axis in a Vega-Lite plot."""
    _method = "line"

class MarkerChannel(Terminal1dChannel):
    """Represents a marker axis in a Vega-Lite plot."""
    _method = "marker"

class Terminal2dChannel(Channel):
    """Abstract class for the last graphical channel in a 2d Vega-Lite plot."""

    def __init__(self, xaxis, yaxis, profile, width, height, title, config, xscale, yscale, colorscale):
        self.xaxis = xaxis
        self.yaxis = yaxis
        self.profile = profile
        self.width = width
        self.height = height
        self.title = title
        self.config = config
        self.xscale = xscale
        self.yscale = yscale
        self.colorscale = colorscale

    def __repr__(self):
        args = [repr(self.xaxis), repr(self.yaxis)]
        if self.profile is not None:
            args.append("profile={0}".format(self.profile))
        if self.width is not None:
            args.append("width={0}".format(repr(self.width)))
        if self.height is not None:
            args.append("height={0}".format(repr(self.height)))
        if self.title is not None:
            args.append("title={0}".format(repr(self.title)))
        if self.config is not None:
            args.append("config={0}".format(repr(self.config)))
        if self.xscale is not None:
            args.append("xscale={0}".format(repr(self.xscale)))
        if self.yscale is not None:
            args.append("yscale={0}".format(repr(self.yscale)))
        if self.colorscale is not None:
            args.append("colorscale={0}".format(repr(self.colorscale)))
        return ".{0}({1})".format(self._method, "".join(args))

class HeatmapChannel(Terminal2dChannel):
    """Represents a heatmap in a Vega-Lite plot."""
    _method = "heatmap"

class PlottingChain(object):
    """Mix-in for :py:class:`Hist <histbook.hist.Hist>` (as the first in a plotting chain) and :py:class:`Channels <histbook.vega.Channel>` in a plotting chain."""

    def __init__(self, source, item):
        if isinstance(source, PlottingChain):
            self._source = source._source
            self._chain = source._chain + (item,)
        else:
            self._source = source
            self._chain = (item,)

    def __repr__(self):
        return "".join(repr(x) for x in (self._source,) + self._chain)

    def __str__(self, indent="\n     ", paren=True):
        return ("(" if paren else "") + indent.join(repr(x) for x in (self._source,) + self._chain) + (")" if paren else "")

    def _singleaxis(self, axis):
        if axis is None:
            if len(self._source._group + self._source._fixed) == 1:
                axis, = self._source._group + self._source._fixed
            else:
                raise TypeError("histogram has more than one axis; one must be specified for plotting")
        return axis

    def _asaxis(self, axis):
        if axis is None:
            return None
        elif isinstance(axis, histbook.axis.Axis):
            return axis
        else:
            return self._source.axis[axis]

    def overlay(self, axis):
        """
        Display bins in ``axis`` overlaid on each other in different colors.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, OverlayChannel) for x in self._chain):
            raise TypeError("cannot overlay an overlay")
        return PlottingChain(self, OverlayChannel(self._asaxis(axis)))

    def stack(self, axis, order=None):
        """
        Display bins in ``axis`` stacked on one another in an area plot.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        order : iterable of strings
            stacking order of bins

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("cannot stack a stack")
        return PlottingChain(self, StackChannel(self._asaxis(axis), order))

    def beside(self, axis):
        """
        Display bins in ``axis`` next to each other horizontally.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, BesideChannel) for x in self._chain):
            raise TypeError("cannot split plots beside each other that are already split with beside (can do beside and below)")
        return PlottingChain(self, BesideChannel(self._asaxis(axis)))

    def below(self, axis):
        """
        Display bins in ``axis`` next to each other vertically.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, BelowChannel) for x in self._chain):
            raise TypeError("cannot split plots below each other that are already split with below (can do beside and below)")
        return PlottingChain(self, BelowChannel(self._asaxis(axis)))

    def bar(self, axis=None, profile=None, error=False, normalized=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as bars on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        normalized : bool
            if ``True``, normalize the histogram

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable1d <histbook.vega.Plotable1d>`
        """
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable1d(self, BarChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def step(self, axis=None, profile=None, error=False, normalized=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as steps on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        normalized : bool
            if ``True``, normalize the histogram

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable1d <histbook.vega.Plotable1d>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable1d(self, StepChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def area(self, axis=None, profile=None, error=False, normalized=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as areas on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        normalized : bool
            if ``True``, normalize the histogram

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable1d <histbook.vega.Plotable1d>`
        """
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable1d(self, AreaChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def line(self, axis=None, profile=None, error=False, normalized=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as lines on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        normalized : bool
            if ``True``, normalize the histogram

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable1d <histbook.vega.Plotable1d>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable1d(self, LineChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def marker(self, axis=None, profile=None, error=True, normalized=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as markers on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        normalized : bool
            if ``True``, normalize the histogram

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable1d <histbook.vega.Plotable1d>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable1d(self, MarkerChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, normalized, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def heatmap(self, xaxis=None, yaxis=None, profile=None, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None):
        """
        Display bins in ``xaxis`` and ``yaxis`` (if not the only two axes) as a heatmap.

        Parameters
        ----------
        xaxis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the horizontal axis to overlay; if ``None`` *(default)*, use the first of the only two axes in this :py:class:`Hist <histbook.hist.Hist>`

        yaxis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the vertical axis to overlay; if ``None`` *(default)*, use the second of the only two axes in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        width, height, title, config, xscale, yscale, colorscale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable2d <histbook.vega.Plotable2d>`
        """
        if any(isinstance(x, OverlayChannel) for x in self._chain):
            raise TypeError("two dimensional plots can't be overlaid")
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("two dimensional plots can't be stacked")

        if xaxis is None and yaxis is None:
            if len(self._source._group + self._source._fixed) == 2:
                xaxis, yaxis = self._source._group + self._source._fixed
            else:
                raise TypeError("histogram doesn't have exactly two axes; an x and y must be specified for plotting")
        elif xaxis is None or yaxis is None:
            raise TypeError("xaxis and yaxis must both be None or neither be None")

        return Plotable2d(self, HeatmapChannel(self._asaxis(xaxis), self._asaxis(yaxis), self._asaxis(profile), width, height, title, config, xscale, yscale, colorscale))

class PlotableFrontends(object):
    def to(self, fcn):
        """Call ``fcn`` on the Vega-Lite JSON for this plot."""
        return fcn(self.vegalite())

    def ipyvega(self):
        """Draw this plot inline in a Jupyter notebook (not lab) using the vega library."""
        import vega
        return self.to(vega.VegaLite)

    def vegascope(self, canvas=None):
        """Draw this plot in a (possibly remote) browser using the VegaScope library."""
        import vegascope
        if canvas is None:
            if not hasattr(vegascope, "canvas"):
                vegascope.canvas = vegascope.LocalCanvas()
            canvas = vegascope.canvas
        return self.to(canvas)

    def altair(self, validate=True):
        """Return an altair.Chart of this plot using the Altair library."""
        import altair
        return self.to(lambda x: altair.Chart.from_dict(x, validate=validate))

    def _repr_mimebundle_(self, include=None, exclude=None, **kwargs):
        return {"application/vnd.vegalite.v2+json": self.vegalite()}

class Plotable1d(PlotableFrontends):
    """Mix-in for :py:class:`Channels <histbook.vega.Channel>` and :py:class:`Combinations <histbook.vega.Combination>` that can be plotted."""

    def __init__(self, source, item):
        if isinstance(source, PlottingChain):
            self._source = source._source
            self._chain = source._chain + (item,)
        else:
            self._source = source
            self._chain = (item,)

    def __repr__(self):
        return "".join(repr(x) for x in (self._source,) + self._chain)

    def __str__(self, indent="\n     ", paren=True):
        return ("(" if paren else "") + indent.join(repr(x) for x in (self._source,) + self._chain) + (")" if paren else "")

    @property
    def _last(self):
        return self._chain[-1]

    def _data(self, prefix, varname):
        error = self._last.error
        baseline = isinstance(self._last, (StepChannel, AreaChannel))

        if isinstance(self._last.axis, (histbook.axis.bin, histbook.axis.intbin, histbook.axis.split)):
            if isinstance(self._last, BarChannel):
                xtype = "ordinal"
            else:
                xtype = "quantitative"
        else:
            xtype = "nominal"

        profile = self._last.profile
        if profile is None:
            profiles = ()
        else:
            profiles = (profile,)

        projected = self._source.project(*(x.axis for x in self._chain))

        table = projected.table(*profiles, count=(profile is None), error=error, normalized=self._last.normalized, recarray=False)

        projectedorder = [x for x in projected.axis if not isinstance(x, histbook.axis.ProfileAxis)]
        lastj = projectedorder.index(self._last.axis)

        data = []
        domains = {}
        logrange = None
        if error and baseline:
            shifts = {}

        def recurse(j, content, row, base):
            if j == len(projectedorder):
                if base:
                    row = row + ((0.0, 0.0) if error else (0.0,))
                else:
                    row = row + tuple(float(x) for x in content)

                datum = dict(prefix + tuple(zip([varname + str(i) for i in range(len(row))], row)))
                data.append(datum)

            else:
                axis = projectedorder[j]
                if axis not in domains:
                    if isinstance(axis, histbook.axis.groupby) and axis.keeporder:
                        domains[axis] = []
                    else:
                        domains[axis] = set()

                if isinstance(domains[axis], list):
                    domains[axis].extend(axis.keys(content))
                else:
                    domains[axis].update(axis.keys(content))

                if isinstance(axis, histbook.axis.intbin):
                    axis = axis.bin()

                for i, (n, x) in enumerate(axis.items(content)):
                    if isinstance(n, histbook.axis.Interval):
                        if j == lastj and xtype == "quantitative":
                            if numpy.isfinite(n.low) and numpy.isfinite(n.high):
                                low = n.low
                                if baseline and isinstance(axis, (histbook.axis.bin, histbook.axis.split)) and n.low == axis.low:
                                    recurse(j + 1, x, row + (n.low,), True)
                                    low += 1e-10*(axis.high - axis.low)
                                elif not baseline and isinstance(axis, (histbook.axis.bin, histbook.axis.split)):
                                    low = 0.5*(n.low + n.high)

                                recurse(j + 1, x, row + (low,), base)

                                if error and baseline:
                                    shifts[low] = 0.5*(n.low + n.high)

                                if baseline and isinstance(axis, (histbook.axis.bin, histbook.axis.split)) and n.high == axis.high:
                                    recurse(j + 1, x, row + (n.high,), True)

                        else:
                            recurse(j + 1, x, row + (str(n),), base)

                    elif isinstance(n, (bool, numpy.bool_, numpy.bool)):
                        recurse(j + 1, x, row + ("pass" if n else "fail",), base)

                    else:
                        recurse(j + 1, x, row + (str(n),), base)

        recurse(0, table, (), False)
        if error and baseline:
            for x in data:
                x[varname + str(lastj) + "c"] = shifts.get(x[varname + str(lastj)], x[varname + str(lastj)])

        if self._last.yscale == "log" or (isinstance(self._last.yscale, dict) and self._last.yscale.get("type") == "log"):
            try:
                logrange = (min(x[varname + str(len(projectedorder))] for x in data if x[varname + str(len(projectedorder))] > 0),
                            max(x[varname + str(len(projectedorder))] for x in data if x[varname + str(len(projectedorder))] > 0))
            except ValueError:
                logrange = (0.5, 2.0)

            if logrange[0] == logrange[1]:
                logrange = (0.5*logrange[0], 2*logrange[0])

            if isinstance(self._last, (LineChannel, MarkerChannel)):
                fillvalue = None    # means "do not draw"
            else:
                fillvalue = 1e-45   # smallest positive float32 value

            for x in data:
                if x[varname + str(len(projectedorder))] <= 0:
                    x[varname + str(len(projectedorder))] = fillvalue

        return projectedorder, data, domains, logrange

    def _vegalite(self, axis, domains, logrange, varname):
        error = self._last.error
        baseline = isinstance(self._last, (StepChannel, AreaChannel))
        if isinstance(self._last.axis, (histbook.axis.bin, histbook.axis.intbin, histbook.axis.split)):
            if isinstance(self._last, BarChannel):
                xtype = "ordinal"
            else:
                xtype = "quantitative"
        else:
            xtype = "nominal"

        if isinstance(self._last, BarChannel):
            mark = {"type": "bar", "clip": True}
        elif isinstance(self._last, StepChannel):
            if xtype == "nominal":
                mark = {"type": "bar", "clip": True}
            else:
                mark = {"type": "line", "interpolate": "step-before", "clip": True}
        elif isinstance(self._last, AreaChannel):
            if xtype == "nominal":
                mark = {"type": "bar", "clip": True}
            else:
                mark = {"type": "area", "interpolate": "step-before", "clip": True}
        elif isinstance(self._last, LineChannel):
            mark = {"type": "line", "clip": True}
        elif isinstance(self._last, MarkerChannel):
            mark = {"type": "point", "clip": True}
        else:
            raise AssertionError(self._last)

        # FIXME: let the user override this with explicit xtitle, ytitle parameters
        xtitle = self._last.axis.expr
        if self._last.profile is None:
            if self._last.normalized:
                ytitle = "probability per bin"
            else:
                ytitle = "entries per bin"
        else:
            ytitle = self._last.profile.expr

        transform = []
        def makeorder(i, var, values):
            if len(values) == 1:
                return "if(datum.{0} === {1}, {2}, {3})".format(var, repr(values[0]), i, i + 1)
            elif len(values) > 1:
                return "if(datum.{0} === {1}, {2}, {3})".format(var, repr(values[0]), i, makeorder(i + 1, var, values[1:]))
            else:
                raise AssertionError(values)

        encoding = {"x": {"field": varname + str(axis.index(self._last.axis)), "type": xtype, "scale": {"zero": False}, "axis": {"title": xtitle}},
                    "y": {"field": varname + str(len(axis)), "type": "quantitative", "axis": {"title": ytitle}}}
        for channel in self._chain[:-1]:
            if isinstance(channel, OverlayChannel):
                overlayorder = [str(x) for x in sorted(domains[channel.axis])]
                encoding["color"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "legend": {"title": channel.axis.expr}, "scale": {"domain": overlayorder}}

            elif isinstance(channel, StackChannel):
                if channel.order is not None:
                    stackorder = channel.order
                elif isinstance(domains[channel.axis], list):
                    stackorder = [str(x) for x in reversed(domains[channel.axis])]
                else:
                    stackorder = [str(x) for x in sorted(domains[channel.axis])]

                encoding["color"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "legend": {"title": channel.axis.expr}, "scale": {"domain": list(reversed(stackorder))}}
                encoding["y"]["aggregate"] = "sum"
                encoding["order"] = {"field": "stackorder", "type": "nominal"}
                transform.append({"calculate": makeorder(0, varname + str(axis.index(channel.axis)), stackorder), "as": "stackorder"})

            elif isinstance(channel, BesideChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["column"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            elif isinstance(channel, BelowChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["row"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            else:
                raise AssertionError(channel)

        if self._last.xscale is not None and "x" in encoding:
            if isinstance(self._last.xscale, dict):
                encoding["x"]["scale"] = copy.deepcopy(self._last.xscale)
            else:
                encoding["x"]["scale"] = {"type": copy.deepcopy(self._last.xscale)}
        if self._last.yscale is not None and "y" in encoding:
            if isinstance(self._last.yscale, dict):
                encoding["y"]["scale"] = copy.deepcopy(self._last.yscale)
            else:
                encoding["y"]["scale"] = {"type": copy.deepcopy(self._last.yscale)}
            if encoding["y"]["scale"].get("type") == "log" and logrange is not None and "domain" not in encoding["y"]["scale"]:
                encoding["y"]["scale"]["domain"] = logrange
        if self._last.colorscale is not None and "color" in encoding:
            if isinstance(self._last.colorscale, dict):
                encoding["color"]["scale"] = copy.deepcopy(self._last.colorscale)
            else:
                encoding["color"]["scale"] = {"type": copy.deepcopy(self._last.colorscale)}
        if self._last.shapescale is not None and "shape" in encoding:
            if isinstance(self._last.shapescale, dict):
                encoding["shape"]["scale"] = copy.deepcopy(self._last.shapescale)
            else:
                encoding["shape"]["scale"] = {"type": copy.deepcopy(self._last.shapescale)}

        if not error:
            return [mark], [encoding], [transform]

        else:
            if error and baseline:
                errx = varname + str(axis.index(self._last.axis)) + "c"
            else:
                errx = varname + str(axis.index(self._last.axis))

            encoding2 = {"x": {"field": errx, "type": "quantitative"},
                         "y": {"field": "error-down", "type": "quantitative"},
                         "y2": {"field": "error-up", "type": "quantitative"}}

            transform2 = [{"calculate": "datum.{0} - datum.{1}".format(varname + str(len(axis)), varname + str(len(axis) + 1)), "as": "error-down"},
                          {"calculate": "datum.{0} + datum.{1}".format(varname + str(len(axis)), varname + str(len(axis) + 1)), "as": "error-up"}]

            return [mark, "rule"], [encoding, encoding2], [transform, transform2]
        
    def vegalite(self):
        """Return the Vega-Lite JSON for this plot."""

        axis, data, domains, logrange = self._data((), "a")
        marks, encodings, transforms = self._vegalite(axis, domains, logrange, "a")

        if len(marks) == 1:
            return self._options({"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
                                  "data": {"values": data},
                                  "mark": marks[0],
                                  "encoding": encodings[0],
                                  "transform": transforms[0]})

        else:
            return self._options({"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
                                  "data": {"values": data},
                                  "layer": [{"mark": m, "encoding": e, "transform": t} for m, e, t in zip(marks, encodings, transforms)]})

    def _options(self, out, config=True):
        if self._last.width is not None:
            out["width"] = copy.deepcopy(self._last.width)
        if self._last.height is not None:
            out["height"] = copy.deepcopy(self._last.height)
        if self._last.title is not None:
            out["title"] = copy.deepcopy(self._last.title)
        if config and self._last.config is not None:
            out["config"] = copy.deepcopy(self._last.config)
        return out

class Plotable2d(PlotableFrontends):
    """Mix-in for :py:class:`Channels <histbook.vega.Channel>` and :py:class:`Combinations <histbook.vega.Combination>` that can be plotted."""

    def __init__(self, source, item):
        if isinstance(source, PlottingChain):
            self._source = source._source
            self._chain = source._chain + (item,)
        else:
            self._source = source
            self._chain = (item,)

    def __repr__(self):
        return "".join(repr(x) for x in (self._source,) + self._chain)

    def __str__(self, indent="\n     ", paren=True):
        return ("(" if paren else "") + indent.join(repr(x) for x in (self._source,) + self._chain) + (")" if paren else "")

    @property
    def _last(self):
        return self._chain[-1]

    def _data(self, prefix, varname):
        profile = self._last.profile
        if profile is None:
            profiles = ()
        else:
            profiles = (profile,)

        projected = self._source.project(*([x.axis for x in self._chain[:-1]] + [self._last.xaxis, self._last.yaxis]))
        table = projected.table(*profiles, count=(profile is None), error=False, recarray=False)

        projectedorder = [x for x in projected.axis if not isinstance(x, histbook.axis.ProfileAxis)]
        lastj = (projectedorder.index(self._last.xaxis), projectedorder.index(self._last.yaxis))
        
        data = []
        domains = {}
        logrange = None

        def recurse(j, content, row):
            if j == len(projectedorder):
                row = row + tuple(float(x) for x in content)

                datum = dict(prefix + tuple(zip([varname + str(i) for i in range(len(row))], row)))
                # FIXME: protect against logarithmic x and y axes
                data.append(datum)

            else:
                axis = projectedorder[j]
                if axis not in domains:
                    domains[axis] = set()
                domains[axis].update(axis.keys(content))

                if isinstance(axis, histbook.axis.intbin):
                    axis = axis.bin()

                for i, (n, x) in enumerate(axis.items(content)):
                    if isinstance(n, histbook.axis.Interval):
                        if j in lastj:
                            if numpy.isfinite(n.low) and numpy.isfinite(n.high):
                                recurse(j + 1, x, row + (0.5*(n.low + n.high),))
                        else:
                            recurse(j + 1, x, row + (str(n),))

                    elif isinstance(n, (bool, numpy.bool_, numpy.bool)):
                        recurse(j + 1, x, row + ("pass" if n else "fail",))

                    else:
                        recurse(j + 1, x, row + (str(n),))

        recurse(0, table, ())
        return projectedorder, data, domains, logrange

    def _vegalite(self, axis, domains, logrange, varname):
        if isinstance(self._last.xaxis, histbook.axis.bin):
            xtype = "quantitative"
            xbin = {"extent": [self._last.xaxis.low, self._last.xaxis.high], "minstep": self._last.xaxis.binwidth, "maxbins": self._last.xaxis.numbins + 1, "nice": False}
        elif isinstance(self._last.xaxis, histbook.axis.intbin):
            xtype = "quantitative"
            xbin = {"extent": [self._last.xaxis.min - 0.5, self._last.xaxis.max + 0.5], "minstep": self._last.xaxis.binwidth, "maxbins": self._last.xaxis.numbins + 1, "nice": False}
        elif isinstance(self._last.xaxis, histbook.axis.split):
            raise NotImplementedError("Vega-Lite doesn't support heatmaps of explicit axis splitting (yet?)")
        else:
            xtype = "nominal"
            xbin = False

        if isinstance(self._last.yaxis, histbook.axis.bin):
            ytype = "quantitative"
            ybin = {"extent": [self._last.yaxis.low, self._last.yaxis.high], "minstep": self._last.yaxis.binwidth, "maxbins": self._last.yaxis.numbins + 1, "nice": False}
        elif isinstance(self._last.yaxis, histbook.axis.intbin):
            ytype = "quantitative"
            ybin = {"extent": [self._last.yaxis.min - 0.5, self._last.yaxis.may + 0.5], "minstep": self._last.yaxis.binwidth, "maxbins": self._last.yaxis.numbins + 1, "nice": False}
        elif isinstance(self._last.yaxis, histbook.axis.split):
            raise NotImplementedError("Vega-Lite doesn't support heatmaps of explicit axis splitting (yet?)")
        else:
            ytype = "nominal"
            ybin = False

        # FIXME: let the user override this with explicit xtitle, ytitle, colortitle parameters
        xtitle = self._last.xaxis.expr
        ytitle = self._last.yaxis.expr
        if self._last.profile is None:
            colortitle = "entries per bin"
            colorzero = True
        else:
            colortitle = self._last.profile.expr
            colorzero = False

        encoding = {"x": {"field": varname + str(axis.index(self._last.xaxis)), "type": xtype, "scale": {"zero": False}, "axis": {"title": xtitle}, "bin": xbin},
                    "y": {"field": varname + str(axis.index(self._last.yaxis)), "type": ytype, "scale": {"zero": False}, "axis": {"title": ytitle}, "bin": ybin},
                    "color": {"field": varname + str(len(axis)), "type": "quantitative", "legend": {"title": colortitle}, "scale": {"zero": colorzero}}}
        for channel in self._chain[:-1]:
            if isinstance(channel, BesideChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["column"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            elif isinstance(channel, BelowChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["row"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            else:
                raise AssertionError(channel)

        if self._last.xscale is not None and "x" in encoding:
            if isinstance(self._last.xscale, dict):
                encoding["x"]["scale"] = copy.deepcopy(self._last.xscale)
            else:
                encoding["x"]["scale"] = {"type": copy.deepcopy(self._last.xscale)}
        if self._last.yscale is not None and "y" in encoding:
            if isinstance(self._last.yscale, dict):
                encoding["y"]["scale"] = copy.deepcopy(self._last.yscale)
            else:
                encoding["y"]["scale"] = {"type": copy.deepcopy(self._last.yscale)}
        if self._last.colorscale is not None and "color" in encoding:
            if isinstance(self._last.colorscale, dict):
                encoding["color"]["scale"] = copy.deepcopy(self._last.colorscale)
            else:
                encoding["color"]["scale"] = {"type": copy.deepcopy(self._last.colorscale)}

        return ["rect"], [encoding], [[]]

    def vegalite(self):
        """Return the Vega-Lite JSON for this plot."""

        axis, data, domains, logrange = self._data((), "a")
        marks, encodings, transforms = self._vegalite(axis, domains, logrange, "a")

        return self._options({"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
                              "data": {"values": data},
                              "mark": marks[0],
                              "encoding": encodings[0],
                              "transform": transforms[0]})

    def _options(self, out, config=True):
        if self._last.width is not None:
            out["width"] = copy.deepcopy(self._last.width)
        if self._last.height is not None:
            out["height"] = copy.deepcopy(self._last.height)
        if self._last.title is not None:
            out["title"] = copy.deepcopy(self._last.title)
        if config and self._last.config is not None:
            out["config"] = copy.deepcopy(self._last.config)
        return out

class Combination(PlotableFrontends):
    """Abstract class for :py:class:`Plotable1ds <histbook.vega.Plotable1d>` that have been combined as an overlay or side-by-side plots."""

    def __init__(self, plotables, types, opts):
        self._plotables = []
        for arg in plotables:           # first level: for varargs
            try:
                iter(arg)
            except:
                arg = [arg]
            for plotable in arg:        # second level: for iterators
                if not isinstance(plotable, types):
                    raise TypeError("only Plotable1ds can be combined with {0}".format(self.__class__.__name__))
                self._plotables.append(plotable)

        self.config = opts.pop("config", None)
        if len(opts) > 0:
            raise TypeError("unrecognized options for {0}: {1}".format(self.__class__.__name__, " ".join(opts)))

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, ", ".join(repr(x) for x in self._plotables))

    def __str__(self, indent="\n    ", paren=False):
        return "{0}({1})".format(self.__class__.__name__, "".join(indent + x.__str__(indent + "    ", False) for x in self._plotables))

    def _varname(self, i):
        out = []
        i += 1
        while i > 0:
            out.append("abcdefghijklmnopqrstuvwxyz"[i % 26])
            i //= 26
        return "".join(reversed(out))

    def _collectdata(self):
        allaxis = []
        alldata = []
        alldomains = []
        alllogrange = []

        def recurse(i, plotables):
            for plotable in plotables:
                if isinstance(plotable, Combination):
                    i = recurse(i, plotable._plotables)
                else:
                    varname = self._varname(i)
                    axis, data, domains, logrange = plotable._data((("id", varname),), varname)
                    allaxis.append(axis)
                    alldata.extend(data)
                    alldomains.append(domains)
                    alllogrange.append(logrange)
                    i += 1
            return i

        recurse(0, self._plotables)
        return allaxis, alldata, alldomains, alllogrange

    def _options(self, out):
        config = {}

        def recurse(node):
            if isinstance(node, Combination):
                for x in node._plotables:
                    recurse(x)
            elif node._last.config is not None:
                config.update(node._last.config)

        recurse(self)

        if self.config is not None:
            config.update(self.config)
        
        if len(config) != 0:
            out["config"] = config

        return out

class overlay(Combination):
    """:py:class:`Plotable1d <histbook.vega.Plotable1d>` overlaying two or more independently produced :py:class:`Plotable1ds <histbook.vega.Plotable1d>`."""

    def __init__(self, *plotables, **opts):
        super(overlay, self).__init__(plotables, (Plotable1d,), opts)

    def _fill(self, i, allaxis, alldomains, alllogrange, tofill):
        for plotable in self._plotables:
            varname = self._varname(i)
            marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], alllogrange[i], varname)
            thislayer = [{"filter": {"field": "id", "equal": varname}}]

            if len(marks) == 1:
                tofill.append(plotable._options({"mark": marks[0],
                                                 "encoding": encodings[0],
                                                 "transform": transforms[0] + thislayer}, config=False))
            else:
                for m, e, t in zip(marks, encodings, transforms):
                    tofill.append(plotable._options({"mark": m, "encoding": e, "transform": t + thislayer}, config=False))
            i += 1

        return i

    def vegalite(self):
        allaxis, alldata, alldomains, alllogrange = self._collectdata()

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "layer": []}

        self._fill(0, allaxis, alldomains, alllogrange, out["layer"])
        return self._options(out)

class beside(Combination):
    """:py:class:`Plotable1d <histbook.vega.Plotable1d>` displaying two or more independently produced :py:class:`Plotable1ds <histbook.vega.Plotable1d>` beside each other horizontally."""

    def __init__(self, *plotables, **opts):
        super(beside, self).__init__(plotables, (Plotable1d, Plotable2d, overlay, below), opts)
        if any(isinstance(x, BesideChannel) for x in self._plotables):
            raise TypeError("cannot place plots beside each other that are already split with beside (can do beside and below)")

    def _fill(self, i, allaxis, alldomains, alllogrange, tofill):
        for plotable in self._plotables:
            if isinstance(plotable, overlay):
                tofill.append({"layer": []})
                i = plotable._fill(i, allaxis, alldomains, alllogrange, tofill[-1]["layer"])

            elif isinstance(plotable, below):
                tofill.append({"vconcat": []})
                i = plotable._fill(i, allaxis, alldomains, alllogrange, tofill[-1]["vconcat"])

            else:
                varname = self._varname(i)
                marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], alllogrange[i], varname)
                thislayer = [{"filter": {"field": "id", "equal": varname}}]

                if len(marks) == 1:
                    tofill.append(plotable._options({"mark": marks[0], "encoding": encodings[0], "transform": transforms[0] + thislayer}, config=False))
                else:
                    tofill.append({"layer": [plotable._options({"mark": m, "encoding": e, "transform": t}, config=False) for m, e, t in zip(marks, encodings, transforms)]})
                i += 1

        return i

    def vegalite(self):
        allaxis, alldata, alldomains, alllogrange = self._collectdata()

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "hconcat": []}

        self._fill(0, allaxis, alldomains, alllogrange, out["hconcat"])
        return self._options(out)

class below(Combination):
    """:py:class:`Plotable1d <histbook.vega.Plotable1d>` displaying two or more independently produced :py:class:`Plotable1ds <histbook.vega.Plotable1d>` below each other vertically."""

    def __init__(self, *plotables, **opts):
        super(below, self).__init__(plotables, (Plotable1d, Plotable2d, overlay, beside), opts)
        if any(isinstance(x, BelowChannel) for x in self._plotables):
            raise TypeError("cannot place plots below each other that are already split with below (can do beside and below)")

    def _fill(self, i, allaxis, alldomains, alllogrange, tofill):
        for plotable in self._plotables:
            if isinstance(plotable, overlay):
                tofill.append({"layer": []})
                i = plotable._fill(i, allaxis, alldomains, alllogrange, tofill[-1]["layer"])

            elif isinstance(plotable, beside):
                tofill.append({"hconcat": []})
                i = plotable._fill(i, allaxis, alldomains, alllogrange, tofill[-1]["hconcat"])

            else:
                varname = self._varname(i)
                marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], alllogrange[i], varname)
                thislayer = [{"filter": {"field": "id", "equal": varname}}]

                if len(marks) == 1:
                    tofill.append(plotable._options({"mark": marks[0], "encoding": encodings[0], "transform": transforms[0] + thislayer}, config=False))
                else:
                    tofill.append({"layer": [plotable._options({"mark": m, "encoding": e, "transform": t}, config=False) for m, e, t in zip(marks, encodings, transforms)]})
                i += 1

        return i

    def vegalite(self):
        allaxis, alldata, alldomains, alllogrange = self._collectdata()

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "vconcat": []}

        self._fill(0, allaxis, alldomains, alllogrange, out["vconcat"])
        return self._options(out)

class grid(Combination):
    """:py:class:`Plotable1d <histbook.vega.Plotable1d>` displaying two or more independently produced :py:class:`Plotable1ds <histbook.vega.Plotable1d>` in a rectangular grid of `numcol` columns."""

    def __init__(self, numcol, *plotables, **opts):
        super(grid, self).__init__(plotables, (Plotable1d, Plotable2d, overlay), opts)
        if any(isinstance(x, BesideChannel) for x in self._plotables) or any(isinstance(x, BelowChannel) for x in self._plotables):
            raise TypeError("cannot place plots in a grid that are already split with beside or below")
        self.numcol = numcol

    def _fill(self, i, allaxis, alldomains, alllogrange, tofill):
        for plotable in self._plotables:
            if len(tofill[-1]["hconcat"]) >= self.numcol:
                tofill.append({"hconcat": []})

            if isinstance(plotable, overlay):
                tofill[-1]["hconcat"].append({"layer": []})
                i = plotable._fill(i, allaxis, alldomains, alllogrange, tofill[-1]["hconcat"][-1]["layer"])

            else:
                varname = self._varname(i)
                marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], alllogrange[i], varname)
                thislayer = [{"filter": {"field": "id", "equal": varname}}]

                if len(marks) == 1:
                    tofill[-1]["hconcat"].append(plotable._options({"mark": marks[0], "encoding": encodings[0], "transform": transforms[0] + thislayer}, config=False))

                else:
                    tofill[-1]["hconcat"].append({"layer": [plotable._options({"mark": m, "encoding": e, "transform": t}, config=False) for m, e, t in zip(marks, encodings, transforms)]})

                i += 1

        return i

    def vegalite(self):
        allaxis, alldata, alldomains, alllogrange = self._collectdata()

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "vconcat": [{"hconcat": []}]}

        self._fill(0, allaxis, alldomains, alllogrange, out["vconcat"])
        return self._options(out)



# [Copied and modified from simonpercivall/astunparse](https://github.com/simonpercivall/astunparse)
# ==================================================================================================
# 
# Copyright (c) 2014, Simon Percivall
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# 
# * Neither the name of AST Unparser nor the names of its contributors
# may be used to endorse or promote products derived from this
# software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# 
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
# 
# 1. This LICENSE AGREEMENT is between the Python Software Foundation
# ("PSF"), and the Individual or Organization ("Licensee") accessing
# and otherwise using this software ("Python") in source or binary
# form and its associated documentation.
# 
# 2. Subject to the terms and conditions of this License Agreement,
# PSF hereby grants Licensee a nonexclusive, royalty-free, world-wide
# license to reproduce, analyze, test, perform and/or display
# publicly, prepare derivative works, distribute, and otherwise use
# Python alone or in any derivative version, provided, however, that
# PSF's License Agreement and PSF's notice of copyright, i.e.,
# "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version
# prepared by Licensee.
# 
# 3. In the event Licensee prepares a derivative work that is based on
# or incorporates Python or any part thereof, and wants to make the
# derivative work available to others as provided herein, then
# Licensee hereby agrees to include in any such work a brief summary
# of the changes made to Python.
# 
# 4. PSF is making Python available to Licensee on an "AS IS" basis.
# PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  BY
# WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY
# REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY
# PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY
# THIRD PARTY RIGHTS.
# 
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
# FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A
# RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY
# DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.
# 
# 6. This License Agreement will automatically terminate upon a
# material breach of its terms and conditions.
# 
# 7. Nothing in this License Agreement shall be deemed to create any
# relationship of agency, partnership, or joint venture between PSF
# and Licensee.  This License Agreement does not grant permission to
# use PSF trademarks or trade name in a trademark sense to endorse or
# promote products or services of Licensee, or any third party.
# 
# 8. By copying, installing or otherwise using Python, Licensee
# agrees to be bound by the terms and conditions of this License
# Agreement.

from __future__ import print_function, unicode_literals
import sys
import ast
import os
import tokenize

def tostring(tree):
    return str(Unparser(tree))

# Large float and imaginary literals get turned into infinities in the AST.
# We unparse those infinities to INFSTR.
INFSTR = "1e" + repr(sys.float_info.max_10_exp + 1)

def interleave(inter, f, seq):
    """Call f on each item in seq, calling inter() in between.
    """
    seq = iter(seq)
    try:
        f(next(seq))
    except StopIteration:
        pass
    else:
        for x in seq:
            inter()
            f(x)

class Unparser(object):
    """Methods in this class recursively traverse an AST and
    output source code for the abstract syntax; original formatting
    is disregarded. """

    def __init__(self, tree):
        """Unparser(tree, file=sys.stdout) -> None.
         Print the source for tree to file."""
        self.f = []
        self.future_imports = []
        self._indent = 0
        self.dispatch(tree)

    def fill(self, text = ""):
        "Indent a piece of text, according to the current indentation level"
        self.f.append("\n"+"    "*self._indent + text)

    def write(self, text):
        "Append a piece of text to the current line."
        self.f.append(text)

    def enter(self):
        "Print ':', and increase the indentation."
        self.write(":")
        self._indent += 1

    def leave(self):
        "Decrease the indentation level."
        self._indent -= 1

    def dispatch(self, tree):
        "Dispatcher function, dispatching tree type T to method _T."
        if isinstance(tree, list):
            for t in tree:
                self.dispatch(t)
            return
        meth = getattr(self, "_"+tree.__class__.__name__)
        meth(tree)

    def __str__(self):
        return "".join(self.f)


    ############### Unparsing methods ######################
    # There should be one method per concrete grammar type #
    # Constructors should be grouped by sum type. Ideally, #
    # this would follow the order in the grammar, but      #
    # currently doesn't.                                   #
    ########################################################

    def _Module(self, tree):
        for stmt in tree.body:
            self.dispatch(stmt)

    def _Interactive(self, tree):
        for stmt in tree.body:
            self.dispatch(stmt)

    def _Expression(self, tree):
        self.dispatch(tree.body)

    # stmt
    def _Expr(self, tree):
        self.fill()
        self.dispatch(tree.value)

    def _Import(self, t):
        self.fill("import ")
        interleave(lambda: self.write(", "), self.dispatch, t.names)

    def _ImportFrom(self, t):
        # A from __future__ import may affect unparsing, so record it.
        if t.module and t.module == '__future__':
            self.future_imports.extend(n.name for n in t.names)

        self.fill("from ")
        self.write("." * t.level)
        if t.module:
            self.write(t.module)
        self.write(" import ")
        interleave(lambda: self.write(", "), self.dispatch, t.names)

    def _Assign(self, t):
        self.fill()
        for target in t.targets:
            self.dispatch(target)
            self.write(" = ")
        self.dispatch(t.value)

    def _AugAssign(self, t):
        self.fill()
        self.dispatch(t.target)
        self.write(" "+self.binop[t.op.__class__.__name__]+"= ")
        self.dispatch(t.value)

    def _AnnAssign(self, t):
        self.fill()
        if not t.simple:
            self.write("(")
        self.dispatch(t.target)
        if not t.simple:
            self.write(")")
        self.write(": ")
        self.dispatch(t.annotation)
        if t.value:
            self.write(" = ")
            self.dispatch(t.value)

    def _Return(self, t):
        self.fill("return")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)

    def _Pass(self, t):
        self.fill("pass")

    def _Break(self, t):
        self.fill("break")

    def _Continue(self, t):
        self.fill("continue")

    def _Delete(self, t):
        self.fill("del ")
        interleave(lambda: self.write(", "), self.dispatch, t.targets)

    def _Assert(self, t):
        self.fill("assert ")
        self.dispatch(t.test)
        if t.msg:
            self.write(", ")
            self.dispatch(t.msg)

    def _Exec(self, t):
        self.fill("exec ")
        self.dispatch(t.body)
        if t.globals:
            self.write(" in ")
            self.dispatch(t.globals)
        if t.locals:
            self.write(", ")
            self.dispatch(t.locals)

    def _Print(self, t):
        self.fill("print ")
        do_comma = False
        if t.dest:
            self.write(">>")
            self.dispatch(t.dest)
            do_comma = True
        for e in t.values:
            if do_comma:self.write(", ")
            else:do_comma=True
            self.dispatch(e)
        if not t.nl:
            self.write(",")

    def _Global(self, t):
        self.fill("global ")
        interleave(lambda: self.write(", "), self.write, t.names)

    def _Nonlocal(self, t):
        self.fill("nonlocal ")
        interleave(lambda: self.write(", "), self.write, t.names)

    def _Yield(self, t):
        self.write("(")
        self.write("yield")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")

    def _YieldFrom(self, t):
        self.write("(")
        self.write("yield from")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")

    def _Raise(self, t):
        self.fill("raise")
        if sys.version_info[0] >= 3:
            if not t.exc:
                assert not t.cause
                return
            self.write(" ")
            self.dispatch(t.exc)
            if t.cause:
                self.write(" from ")
                self.dispatch(t.cause)
        else:
            self.write(" ")
            if t.type:
                self.dispatch(t.type)
            if t.inst:
                self.write(", ")
                self.dispatch(t.inst)
            if t.tback:
                self.write(", ")
                self.dispatch(t.tback)

    def _Try(self, t):
        self.fill("try")
        self.enter()
        self.dispatch(t.body)
        self.leave()
        for ex in t.handlers:
            self.dispatch(ex)
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()
        if t.finalbody:
            self.fill("finally")
            self.enter()
            self.dispatch(t.finalbody)
            self.leave()

    def _TryExcept(self, t):
        self.fill("try")
        self.enter()
        self.dispatch(t.body)
        self.leave()

        for ex in t.handlers:
            self.dispatch(ex)
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _TryFinally(self, t):
        if len(t.body) == 1 and isinstance(t.body[0], ast.TryExcept):
            # try-except-finally
            self.dispatch(t.body)
        else:
            self.fill("try")
            self.enter()
            self.dispatch(t.body)
            self.leave()

        self.fill("finally")
        self.enter()
        self.dispatch(t.finalbody)
        self.leave()

    def _ExceptHandler(self, t):
        self.fill("except")
        if t.type:
            self.write(" ")
            self.dispatch(t.type)
        if t.name:
            self.write(" as ")
            if sys.version_info[0] >= 3:
                self.write(t.name)
            else:
                self.dispatch(t.name)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _ClassDef(self, t):
        self.write("\n")
        for deco in t.decorator_list:
            self.fill("@")
            self.dispatch(deco)
        self.fill("class "+t.name)
        if sys.version_info[0] >= 3:
            self.write("(")
            comma = False
            for e in t.bases:
                if comma: self.write(", ")
                else: comma = True
                self.dispatch(e)
            for e in t.keywords:
                if comma: self.write(", ")
                else: comma = True
                self.dispatch(e)
            if sys.version_info[:2] < (3, 5):
                if t.starargs:
                    if comma: self.write(", ")
                    else: comma = True
                    self.write("*")
                    self.dispatch(t.starargs)
                if t.kwargs:
                    if comma: self.write(", ")
                    else: comma = True
                    self.write("**")
                    self.dispatch(t.kwargs)
            self.write(")")
        elif t.bases:
                self.write("(")
                for a in t.bases:
                    self.dispatch(a)
                    self.write(", ")
                self.write(")")
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _generic_FunctionDef(self, t, async_=False):
        self.write("\n")
        for deco in t.decorator_list:
            self.fill("@")
            self.dispatch(deco)
        self.fill(("async " if async_ else "") + "def " + t.name + "(")
        self.dispatch(t.args)
        self.write(")")
        if getattr(t, "returns", False):
            self.write(" -> ")
            self.dispatch(t.returns)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _FunctionDef(self, t):
        self._generic_FunctionDef(t)

    def _AsyncFunctionDef(self, t):
        self._generic_FunctionDef(t, async_=True)

    def _generic_For(self, t, async_=False):
        self.fill("async for " if async_ else "for ")
        self.dispatch(t.target)
        self.write(" in ")
        self.dispatch(t.iter)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _For(self, t):
        self._generic_For(t)

    def _AsyncFor(self, t):
        self._generic_For(t, async_=True)

    def _If(self, t):
        self.fill("if ")
        self.dispatch(t.test)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        # collapse nested ifs into equivalent elifs.
        while (t.orelse and len(t.orelse) == 1 and
               isinstance(t.orelse[0], ast.If)):
            t = t.orelse[0]
            self.fill("elif ")
            self.dispatch(t.test)
            self.enter()
            self.dispatch(t.body)
            self.leave()
        # final else
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _While(self, t):
        self.fill("while ")
        self.dispatch(t.test)
        self.enter()
        self.dispatch(t.body)
        self.leave()
        if t.orelse:
            self.fill("else")
            self.enter()
            self.dispatch(t.orelse)
            self.leave()

    def _generic_With(self, t, async_=False):
        self.fill("async with " if async_ else "with ")
        if hasattr(t, 'items'):
            interleave(lambda: self.write(", "), self.dispatch, t.items)
        else:
            self.dispatch(t.context_expr)
            if t.optional_vars:
                self.write(" as ")
                self.dispatch(t.optional_vars)
        self.enter()
        self.dispatch(t.body)
        self.leave()

    def _With(self, t):
        self._generic_With(t)

    def _AsyncWith(self, t):
        self._generic_With(t, async_=True)

    # expr
    def _Bytes(self, t):
        self.write(repr(t.s))

    def _Str(self, tree):
        if sys.version_info[0] >= 3:
            self.write(repr(tree.s))
        else:
            # if from __future__ import unicode_literals is in effect,
            # then we want to output string literals using a 'b' prefix
            # and unicode literals with no prefix.
            if "unicode_literals" not in self.future_imports:
                self.write(repr(tree.s))
            elif isinstance(tree.s, str):
                self.write("b" + repr(tree.s))
            elif isinstance(tree.s, unicode):
                self.write(repr(tree.s).lstrip("u"))
            else:
                assert False, "shouldn't get here"

    format_conversions = {97: 'a', 114: 'r', 115: 's'}

    def _FormattedValue(self, t):
        # FormattedValue(expr value, int? conversion, expr? format_spec)
        self.write("{")
        self.dispatch(t.value)
        if t.conversion is not None and t.conversion != -1:
            self.write("!")
            self.write(self.format_conversions[t.conversion])
            #raise NotImplementedError(ast.dump(t, True, True))
        if t.format_spec is not None:
            self.write(":")
            if isinstance(t.format_spec, ast.Str):
                self.write(t.format_spec.s)
            else:
                self.dispatch(t.format_spec)
        self.write("}")

    def _JoinedStr(self, t):
        # JoinedStr(expr* values)
        self.write("f'''")
        for value in t.values:
            if isinstance(value, ast.Str):
                self.write(value.s)
            else:
                self.dispatch(value)
        self.write("'''")

    def _Name(self, t):
        self.write(t.id)

    def _NameConstant(self, t):
        self.write(repr(t.value))

    def _Repr(self, t):
        self.write("`")
        self.dispatch(t.value)
        self.write("`")

    def _Num(self, t):
        repr_n = repr(t.n)
        if sys.version_info[0] >= 3:
            self.write(repr_n.replace("inf", INFSTR))
        else:
            # Parenthesize negative numbers, to avoid turning (-1)**2 into -1**2.
            if repr_n.startswith("-"):
                self.write("(")
            if "inf" in repr_n and repr_n.endswith("*j"):
                repr_n = repr_n.replace("*j", "j")
            # Substitute overflowing decimal literal for AST infinities.
            self.write(repr_n.replace("inf", INFSTR))
            if repr_n.startswith("-"):
                self.write(")")

    def _List(self, t):
        self.write("[")
        interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("]")

    def _ListComp(self, t):
        self.write("[")
        self.dispatch(t.elt)
        for gen in t.generators:
            self.dispatch(gen)
        self.write("]")

    def _GeneratorExp(self, t):
        self.write("(")
        self.dispatch(t.elt)
        for gen in t.generators:
            self.dispatch(gen)
        self.write(")")

    def _SetComp(self, t):
        self.write("{")
        self.dispatch(t.elt)
        for gen in t.generators:
            self.dispatch(gen)
        self.write("}")

    def _DictComp(self, t):
        self.write("{")
        self.dispatch(t.key)
        self.write(": ")
        self.dispatch(t.value)
        for gen in t.generators:
            self.dispatch(gen)
        self.write("}")

    def _comprehension(self, t):
        if getattr(t, 'is_async', False):
            self.write(" async")
        self.write(" for ")
        self.dispatch(t.target)
        self.write(" in ")
        self.dispatch(t.iter)
        for if_clause in t.ifs:
            self.write(" if ")
            self.dispatch(if_clause)

    def _IfExp(self, t):
        self.write("(")
        self.dispatch(t.body)
        self.write(" if ")
        self.dispatch(t.test)
        self.write(" else ")
        self.dispatch(t.orelse)
        self.write(")")

    def _Set(self, t):
        assert(t.elts) # should be at least one element
        self.write("{")
        interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write("}")

    def _Dict(self, t):
        self.write("{")
        def write_pair(pair):
            (k, v) = pair
            self.dispatch(k)
            self.write(": ")
            self.dispatch(v)
            self.write(",")
        self._indent +=1
        self.fill("")
        interleave(lambda: self.fill(""), write_pair, zip(t.keys, t.values))
        self._indent -=1
        self.fill("}")

    def _Tuple(self, t):
        self.write("(")
        if len(t.elts) == 1:
            (elt,) = t.elts
            self.dispatch(elt)
            self.write(",")
        else:
            interleave(lambda: self.write(", "), self.dispatch, t.elts)
        self.write(")")

    unop = {"Invert":"~", "Not": "not", "UAdd":"+", "USub":"-"}
    def _UnaryOp(self, t):
        self.write("(")
        self.write(self.unop[t.op.__class__.__name__])
        self.write(" ")
        if sys.version_info[0] <= 2 and isinstance(t.op, ast.USub) and isinstance(t.operand, ast.Num):
            # If we're applying unary minus to a number, parenthesize the number.
            # This is necessary: -2147483648 is different from -(2147483648) on
            # a 32-bit machine (the first is an int, the second a long), and
            # -7j is different from -(7j).  (The first has real part 0.0, the second
            # has real part -0.0.)
            self.write("(")
            self.dispatch(t.operand)
            self.write(")")
        else:
            self.dispatch(t.operand)
        self.write(")")

    binop = { "Add":"+", "Sub":"-", "Mult":"*", "Div":"/", "Mod":"%",
                    "LShift":"<<", "RShift":">>", "BitOr":"|", "BitXor":"^", "BitAnd":"&",
                    "FloorDiv":"//", "Pow": "**",
                    "MatMult":"@"}
    def _BinOp(self, t):
        self.write("(")
        self.dispatch(t.left)
        self.write(" " + self.binop[t.op.__class__.__name__] + " ")
        self.dispatch(t.right)
        self.write(")")

    cmpops = {"Eq":"==", "NotEq":"!=", "Lt":"<", "LtE":"<=", "Gt":">", "GtE":">=",
                        "Is":"is", "IsNot":"is not", "In":"in", "NotIn":"not in"}
    def _Compare(self, t):
        self.write("(")
        self.dispatch(t.left)
        for o, e in zip(t.ops, t.comparators):
            self.write(" " + self.cmpops[o.__class__.__name__] + " ")
            self.dispatch(e)
        self.write(")")

    boolops = {ast.And: 'and', ast.Or: 'or'}
    def _BoolOp(self, t):
        self.write("(")
        s = " %s " % self.boolops[t.op.__class__]
        interleave(lambda: self.write(s), self.dispatch, t.values)
        self.write(")")

    def _Attribute(self,t):
        self.dispatch(t.value)
        # Special case: 3.__abs__() is a syntax error, so if t.value
        # is an integer literal then we need to either parenthesize
        # it or add an extra space to get 3 .__abs__().
        if isinstance(t.value, ast.Num) and isinstance(t.value.n, int):
            self.write(" ")
        self.write(".")
        self.write(t.attr)

    def _Call(self, t):
        self.dispatch(t.func)
        self.write("(")
        comma = False
        for e in t.args:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        for e in t.keywords:
            if comma: self.write(", ")
            else: comma = True
            self.dispatch(e)
        if sys.version_info[:2] < (3, 5):
            if t.starargs:
                if comma: self.write(", ")
                else: comma = True
                self.write("*")
                self.dispatch(t.starargs)
            if t.kwargs:
                if comma: self.write(", ")
                else: comma = True
                self.write("**")
                self.dispatch(t.kwargs)
        self.write(")")

    def _Subscript(self, t):
        self.dispatch(t.value)
        self.write("[")
        self.dispatch(t.slice)
        self.write("]")

    def _Starred(self, t):
        self.write("*")
        self.dispatch(t.value)

    # slice
    def _Ellipsis(self, t):
        self.write("...")

    def _Index(self, t):
        self.dispatch(t.value)

    def _Slice(self, t):
        if t.lower:
            self.dispatch(t.lower)
        self.write(":")
        if t.upper:
            self.dispatch(t.upper)
        if t.step:
            self.write(":")
            self.dispatch(t.step)

    def _ExtSlice(self, t):
        interleave(lambda: self.write(', '), self.dispatch, t.dims)

    # argument
    def _arg(self, t):
        self.write(t.arg)
        if t.annotation:
            self.write(": ")
            self.dispatch(t.annotation)

    # others
    def _arguments(self, t):
        first = True
        # normal arguments
        defaults = [None] * (len(t.args) - len(t.defaults)) + t.defaults
        for a,d in zip(t.args, defaults):
            if first:first = False
            else: self.write(", ")
            self.dispatch(a)
            if d:
                self.write("=")
                self.dispatch(d)

        # varargs, or bare '*' if no varargs but keyword-only arguments present
        if t.vararg or getattr(t, "kwonlyargs", False):
            if first: first = False
            else: self.write(", ")
            self.write("*")
            if t.vararg:
                if hasattr(t.vararg, 'arg'):
                    self.write(t.vararg.arg)
                    if t.vararg.annotation:
                        self.write(": ")
                        self.dispatch(t.vararg.annotation)
                else:
                    self.write(t.vararg)
                    if getattr(t, 'varargannotation', None):
                        self.write(": ")
                        self.dispatch(t.varargannotation)

        # keyword-only arguments
        if getattr(t, "kwonlyargs", False):
            for a, d in zip(t.kwonlyargs, t.kw_defaults):
                if first:first = False
                else: self.write(", ")
                self.dispatch(a),
                if d:
                    self.write("=")
                    self.dispatch(d)

        # kwargs
        if t.kwarg:
            if first:first = False
            else: self.write(", ")
            if hasattr(t.kwarg, 'arg'):
                self.write("**"+t.kwarg.arg)
                if t.kwarg.annotation:
                    self.write(": ")
                    self.dispatch(t.kwarg.annotation)
            else:
                self.write("**"+t.kwarg)
                if getattr(t, 'kwargannotation', None):
                    self.write(": ")
                    self.dispatch(t.kwargannotation)

    def _keyword(self, t):
        if t.arg is None:
            # starting from Python 3.5 this denotes a kwargs part of the invocation
            self.write("**")
        else:
            self.write(t.arg)
            self.write("=")
        self.dispatch(t.value)

    def _Lambda(self, t):
        self.write("(")
        self.write("lambda ")
        self.dispatch(t.args)
        self.write(": ")
        self.dispatch(t.body)
        self.write(")")

    def _alias(self, t):
        self.write(t.name)
        if t.asname:
            self.write(" as "+t.asname)

    def _withitem(self, t):
        self.dispatch(t.context_expr)
        if t.optional_vars:
            self.write(" as ")
            self.dispatch(t.optional_vars)

    def _Await(self, t):
        self.write("(")
        self.write("await")
        if t.value:
            self.write(" ")
            self.dispatch(t.value)
        self.write(")")



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import math
import numbers

import numpy

import histbook.axis
import histbook.expr

class AxisTuple(tuple):
    """An ordered sequence of :py:class:`Axis <histbook.axis.Axis>` returned by :py:meth:`Hist.axis <histbook.hist.Hist.axis>`."""
    def __getitem__(self, item):
        """Get an axis by algebraic expression (string) or index position (integer)."""
        if isinstance(item, (numbers.Integral, numpy.integer)):
            return super(AxisTuple, self).__getitem__(item)
        else:
            expr = histbook.expr.Expr.parse(item, defs=self._defs)
            for axis in self:
                if expr == axis._parsed:
                    return axis
            raise IndexError("no such axis: {0}".format(repr(str(expr))))

    def _findbyclass(self, expr, cls, kwargs):
        expr = histbook.expr.Expr.parse(expr, defs=self._defs)
        for axis in self:
            if isinstance(axis, cls) and axis._parsed == expr and all(hasattr(axis, n) and getattr(axis, n) == x for n, x in kwargs.items()):
                return axis
        raise IndexError("no such axis: {0}({1}{2})".format(cls.__name__, repr(str(expr)), "".join(", {0}={1}".format(n, kwargs[n]) for n in sorted(kwargs))))

    def groupby(self, expr, **kwargs):
        """Get a :py:class:`groupby <histbook.axis.groupby>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.groupby, kwargs)

    def groupbin(self, expr, **kwargs):
        """Get a :py:class:`groupbin <histbook.axis.groupbin>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.groupbin, kwargs)

    def bin(self, expr, **kwargs):
        """Get a :py:class:`bin <histbook.axis.bin>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.bin, kwargs)

    def intbin(self, expr, **kwargs):
        """Get a :py:class:`intbin <histbook.axis.intbin>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.intbin, kwargs)

    def split(self, expr, **kwargs):
        """Get a :py:class:`split <histbook.axis.split>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.split, kwargs)

    def cut(self, expr, **kwargs):
        """Get a :py:class:`cut <histbook.axis.cut>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.cut, kwargs)

    def profile(self, expr, **kwargs):
        """Get a :py:class:`profile <histbook.axis.profile>` axis by algebraic expression (string) and any provided arguments."""
        return self._findbyclass(expr, histbook.axis.profile, kwargs)

class Projectable(object):
    """Mix-in for :py:class:`Hist <histbook.hist.Hist>` methods that provide selection, projection, and rebinning."""
    @property
    def axis(self):
        """The axes that define a histogram's binning of space."""
        out = AxisTuple(self._group + self._fixed + self._profile)
        out._defs = self._defs
        return out

    def rebin(self, axis, edges):
        """
        Reduce the number of bins by combining existing bins at a specified set of ``edges``.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (string), or index position (integer)
            the axis to rebin

        edges : iterable of numbers
            new bin edges; must be a subset of existing bin edges

        Returns
        -------
        :py:class:`Hist <histbook.hist.Hist>`
            a rebinned histogram.
        """
        if not isinstance(axis, histbook.axis.Axis):
            expr = histbook.expr.Expr.parse(axis, defs=self._defs)
            for x in self._group + self._fixed:
                if isinstance(x, histbook.axis._RebinSplit) and expr == x._parsed:
                    axis = x
                    break

        for index, x in enumerate(self._group + self._fixed):
            if isinstance(x, histbook.axis._RebinSplit) and axis == x:
                axis = x
                break
        else:
            raise IndexError("no such rebinnable axis: {0}".format(axis))

        if isinstance(axis, histbook.axis.GroupAxis):
            newaxis, newcontent = axis._rebinsplit(edges, self._content, index)
        else:
            newaxis, newcontent = axis._rebinsplit(edges, self._content, index - len(self._group))

        outaxis = [newaxis if i == index else x for i, x in enumerate(self._group + self._fixed + self._profile)]
        out = self.__class__(*outaxis, weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        out._content = newcontent
        return out

    def rebinby(self, axis, factor):
        """
        Reduce the number of bins by an approximate ``factor`` by combining existing bins.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (string), or index position (integer)
            the axis to rebin

        factor : positive integer
            number of bins to combine into a single bin (inexact if the number of bins in ``axis`` is not an exact multiple of ``factor``)

        Returns
        -------
        :py:class:`Hist <histbook.hist.Hist>`
            a rebinned histogram.
        """
        if not isinstance(factor, (numbers.Integral, numpy.integer)) or factor <= 0:
            raise TypeError("factor must be a positive integer")

        if not isinstance(axis, histbook.axis.Axis):
            expr = histbook.expr.Expr.parse(axis, defs=self._defs)
            for x in self._group + self._fixed:
                if isinstance(x, histbook.axis._RebinFactor) and expr == x._parsed:
                    axis = x
                    break

        for index, x in enumerate(self._group + self._fixed):
            if isinstance(x, histbook.axis._RebinFactor) and axis == x:
                axis = x
                break
        else:
            raise IndexError("no such rebinnable axis: {0}".format(axis))

        if isinstance(axis, histbook.axis.GroupAxis):
            newaxis, newcontent = axis._rebinsplit(factor, self._content, index)
        else:
            newaxis, newcontent = axis._rebinsplit(factor, self._content, index - len(self._group))

        outaxis = [newaxis if i == index else x for i, x in enumerate(self._group + self._fixed + self._profile)]
        out = self.__class__(*outaxis, weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        out._content = newcontent
        return out

    def drop(self, *profile):
        """
        Remove one or more :py:class:`profile <histbook.axis.profile>` axes.

        Parameters
        ----------
        *profile : :py:class:`profile <histbook.axis.profile>`, algebraic expression (string), or index position (integer)
            the axis or axes to drop

        Returns
        -------
        :py:class:`Hist <histbook.hist.Hist>`
            a histogram without the selected ``profiles``.
        """
        if len(profile) == 0:
            raise IndexError("no profile axis given")
        profile = [x if isinstance(x, histbook.axis.Axis) else self.axis.profile(x) for x in profile]
        for prof in profile:
            if prof not in self._profile:
                raise IndexError("no such profile axis: {0}".format(prof))
        
        axis = []
        index = []
        for i, prof in enumerate(self._profile):
            if prof not in profile:
                axis.append(prof)
                index.append(i)

        index.append(self._sumwindex)
        if self._weightoriginal is not None:
            index.append(self._sumw2index)

        slc = (slice(None),) * (len(self._shape) - 1) + (index,)

        def dropcontent(content):
            if isinstance(content, dict):
                return type(content)((n, dropcontent(x)) for n, x in content.items())
            else:
                return content[slc]

        out = self.__class__(*(self._group + self._fixed + tuple(axis)), weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        if self._content is not None:
            out._content = dropcontent(self._content)
        return out

    def project(self, *axis):
        """
        Project onto a given set of :py:class:`axis <histbook.axis.Axis>`.

        Parameters
        ----------
        *axis : :py:class:`Axis <histbook.axis.axis>`, algebraic expression (string), or index position (integer)
            the axis or axes to keep (all :py:class:`profile <histbook.axis.profile>` axes are kept)

        Returns
        -------
        :py:class:`Hist <histbook.hist.Hist>`
            a histogram projected onto the selected axis or axes.
        """
        allaxis = self._group + self._fixed

        axis = [x if isinstance(x, histbook.axis.Axis) else self.axis[x] for x in axis]
        for x in axis:
            if x not in allaxis:
                raise IndexError("no such axis: {0}".format(x))

        def projarray(content):
            return numpy.sum(content, tuple(i for i, x in enumerate(self._fixed) if x not in axis))

        def addany(left, right):
            if isinstance(left, dict) and isinstance(right, dict):
                if isinstance(left, collections.OrderedDict) or isinstance(right, collections.OrderedDict):
                    dictclass = collections.OrderedDict
                else:
                    dictclass = dict

                out = dictclass((n, addany(x, right[n]) if n in right else x) for n, x in left.items())

                for n, x in right.items():
                    if n not in left:
                        out[n] = x
                return out

            else:
                return left + right

        def addall(values):
            assert len(values) != 0
            left = values[:len(values) // 2]
            right = values[len(values) // 2:]
            if len(left) == 0:
                return right[0]
            elif len(right) == 0:
                return left[0]
            elif len(left) == 1 and len(right) == 1:
                return addany(left[0], right[0])
            else:
                return addall([addall(left)] + [addall(right)])
            
        def projcontent(j, content):
            if j < len(self._group):
                if allaxis[j] in axis:
                    return type(content)((n, projcontent(j + 1, x)) for n, x in content.items())
                else:
                    return addall([projcontent(j + 1, x) for x in content.values()])
            else:
                return projarray(content)

        outaxis = [x for x in allaxis if x in axis] + [x for x in self._profile]

        out = self.__class__(*outaxis, weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))

        if self._content is not None:
            out._content = projcontent(0, self._content)
        return out

    def select(self, expr, tolerance=1e-12):
        """
        Eliminate bins by selecting data with a boolean ``expr``.

        Parameters
        ----------
        expr : algebraic expression (string)
            boolean expression of data to keep; selection thresholds must align with bin edges with the right inequality (e.g. ``<`` vs ``<=``)

        tolerance : small positive number
            absolute difference between selection threshold and bin edge to qualify as a match

        Returns
        -------
        :py:class:`Hist <histbook.hist.Hist>`
            a histogram with data removed (fewer bins)
        """
        expr = histbook.expr.Expr.parse(expr, defs=self._defs)

        if isinstance(expr, histbook.expr.LogicalAnd):
            out = None
            for arg in expr.args:
                if out is None:
                    out = self._select(arg, tolerance)
                else:
                    out = out._select(arg, tolerance)
            return out

        else:
            return self._select(expr, tolerance)

    def _select(self, expr, tolerance):
        if not isinstance(expr, (histbook.expr.Relation, histbook.expr.Logical, histbook.expr.Predicate, histbook.expr.Name)):
            raise TypeError("select expression must be boolean, not {0}".format(repr(str(expr))))

        def asconst(expr):
            if isinstance(expr, histbook.expr.Const):
                return expr

            elif isinstance(expr, histbook.expr.Name) and expr.value in histbook.expr.Expr.maybeconstants:
                return histbook.expr.Const(histbook.expr.Expr.maybeconstants[expr.value])

            elif isinstance(expr, histbook.expr.PlusMinus):
                out = expr.const
                for x in expr.pos:
                    c = asconst(x)
                    if c is None:
                        return None
                    out += c.value
                for x in expr.neg:
                    c = asconst(x)
                    if c is None:
                        return None
                    out -= c.value
                return histbook.expr.Const(out)

            elif isinstance(expr, histbook.expr.TimesDiv):
                out = expr.const
                for x in expr.pos:
                    c = asconst(x)
                    if c is None:
                        return None
                    out *= c.value
                for x in expr.neg:
                    c = asconst(x)
                    if c is None:
                        return None
                    out /= c.value
                return histbook.expr.Const(out)

            else:
                return None

        def normalizeexpr(expr):
            if isinstance(expr, histbook.expr.Relation):
                cutcmp = expr.cmp
                if asconst(expr.left) is not None and asconst(expr.right) is None:
                    cutvalue, cutexpr = asconst(expr.left), expr.right
                    cutcmp = {"<": ">", "<=": ">="}.get(cutcmp, cutcmp)
                elif asconst(expr.left) is None and asconst(expr.right) is not None:
                    cutexpr, cutvalue = expr.left, asconst(expr.right)
                elif asconst(expr.left) is not None and asconst(expr.right) is not None:
                    raise TypeError("select expression must have a variable left or right hand side, not {0}".format(repr(str(expr))))
                else:
                    raise TypeError("select expression must have a constant left or right hand side, not {0}".format(repr(str(expr))))

                cutvalue = cutvalue.value   # unbox to Python

            elif isinstance(expr, histbook.expr.Predicate):
                if expr.positive:
                    cutexpr = histbook.expr.Name(expr.value)
                    cutcmp = "=="
                    cutvalue = True
                else:
                    cutexpr = histbook.expr.Name(expr.value)
                    cutcmp = "!="
                    cutvalue = True

            else:
                cutexpr = expr
                cutcmp = "=="
                cutvalue = True

            return cutexpr, cutcmp, cutvalue

        cutexpr, cutcmp, cutvalue = normalizeexpr(expr)

        allaxis = self._group + self._fixed

        def logical(cutexpr, cutcmp, cutvalue, axis, cutaxis, newaxis, cutslice, closest, closestaxis, wrongcmpaxis):
            if isinstance(axis, (histbook.axis.groupby, histbook.axis.groupbin)) and isinstance(cutexpr, histbook.expr.LogicalOr):
                good = True
                orslice = None
                for andexpr in cutexpr.args:
                    andslice = None
                    for xpr in andexpr.args:
                        xpr, cmp, value = normalizeexpr(xpr)
                        check, _, slice1, _, _, _ = logical(xpr, cmp, value, axis, cutaxis, newaxis, cutslice, closest, closestaxis, wrongcmpaxis)
                        if check is None:
                            good = False
                            break
                        elif andslice is None:
                            andslice = slice1
                        else:
                            slice2 = andslice
                            andslice = lambda x: slice1(x) and slice2(x)

                    if not good:
                        break
                    elif orslice is None:
                        orslice = andslice
                    else:
                        slice1 = andslice
                        slice2 = orslice
                        orslice = lambda x: slice1(x) or slice2(x)

                if good:
                    return axis, axis, orslice, None, None, None

            if isinstance(axis, histbook.axis.cut) and expr == axis._parsed:
                newaxis, cutslice, close, wrongcmp = axis._select("==", True, tolerance)
                if newaxis is not None:
                    return axis, newaxis, cutslice, closest, wrongcmpaxis, closestaxis

            if cutexpr == axis._parsed:
                newaxis, cutslice, close, wrongcmp = axis._select(cutcmp, cutvalue, tolerance)
                if newaxis is not None:
                    cutaxis = axis
                else:
                    if close is not None and (closest is None or abs(cutvalue - close) < abs(cutvalue - closest)):
                        closest = close
                        closestaxis = axis
                    if wrongcmp:
                        wrongcmpaxis = axis

            return cutaxis, newaxis, cutslice, closest, wrongcmpaxis, closestaxis

        cutaxis, newaxis, cutslice, closest, wrongcmpaxis, closestaxis = None, None, None, None, None, None
        for axis in allaxis:
            cutaxis, newaxis, cutslice, closest, wrongcmpaxis, closestaxis = logical(cutexpr, cutcmp, cutvalue, axis, cutaxis, newaxis, cutslice, closest, wrongcmpaxis, closestaxis)
            if cutaxis is not None:
                break
        else:
            if wrongcmpaxis is not None:
                raise ValueError("no axis can select {0} (axis {1} has the wrong inequality; low edges are {2})".format(repr(str(expr)), wrongcmpaxis, "closed" if wrongcmpaxis.closedlow else "open"))
            elif closestaxis is not None:
                raise ValueError("no axis can select {0} (axis {1} has an edge at {2})".format(repr(str(expr)), closestaxis, repr(closest)))
            else:
                raise ValueError("no axis can select {0}".format(repr(str(expr))))

        return self._selectaxis(cutaxis, newaxis, cutslice, True)

    def _selectaxis(self, cutaxis, newaxis, cutslice, dropnull):
        allaxis = self._group + self._fixed

        def cutcontent(i, content):
            if content is None:
                return None

            if isinstance(allaxis[i], (histbook.axis.groupby, histbook.axis.groupbin)):
                if allaxis[i] is cutaxis:
                    return type(content)((n, x) for n, x in content.items() if cutslice(n))
                else:
                    return type(content)((n, cutcontent(i + 1, x)) for n, x in content.items())

            else:
                slc = tuple(cutslice if j < len(allaxis) and allaxis[j] is cutaxis else slice(None) for j in range(i, len(allaxis) + 1))
                out = content[slc].copy()
                if dropnull and isinstance(newaxis, histbook.axis._nullaxis):
                    out.shape = tuple(sh for sh, sl in zip(out.shape, slc) if sl is not cutslice)
                return out

        axis = [newaxis if x is cutaxis else x for x in self._group + self._fixed + self._profile]
        if dropnull:
            axis = [x for x in axis if not isinstance(x, histbook.axis._nullaxis)]
        out = self.__class__(*axis, weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        if self._content is not None:
            out._content = cutcontent(0, self._content)
        return out

    def table(self, *profile, **opts):
        """
        Return histogram data as a table of counts and, optionally, dependent variables (profiles).

        Parameters
        ----------
        *profile : :py:class:`profile <histbook.axis.profile>`
            the dependent variables to include in the table

        Keyword Arguments
        -----------------
        count : bool
            if ``True`` *(default)*, include the (possibly weighted) count of entries in each bin

        effcount : bool
            if ``True`` *(not default)*, include the effective count, which is used to convert between weighted profile errors and weighted profile spreads (equal to ``count`` for unweighted data)

        error : bool
            if ``True`` *(default)*, include "errors" on all parameters (uncertainty in the mean of the distribution the count or profile average represents)

        normalized : bool
            if ``True`` *(not default)*, scale each ``count`` and ``err(count)`` such that the sum over counts times bin widths is 1; does not affect profiles

        recarray : bool
            if ``True`` *(default)*, return results as a Numpy record array, which is rank-2 with named columns; if ``False``, return a plain Numpy array, which is rank-N for N axes and has no column labels.

        columns : bool
            if ``True`` *(not default)*, return a 2-tuple in which the second argument is a list of column labels.
        """
        count = opts.pop("count", True)
        effcount = opts.pop("effcount", False)
        error = opts.pop("error", True)
        normalized = opts.pop("normalized", False)
        recarray = opts.pop("recarray", True)
        showcolumns = opts.pop("columns", False)
        if len(opts) > 0:
            raise TypeError("unrecognized options for Hist.table: {0}".format(" ".join(opts)))

        self._prefill()

        profile = [x if isinstance(x, histbook.axis.Axis) else self.axis.profile(x) for x in profile]
        profileindex = []
        for prof in profile:
            if prof not in self._profile:
                raise IndexError("no such profile axis: {0}".format(prof))
            else:
                profileindex.append(self._profile.index(prof))

        profile = [self._profile[i] for i in profileindex]

        columns = []
        if count:
            columns.append("count()")
            if error:
                columns.append("err(count())")
        if effcount:
            columns.append("effcount()")
        for prof in profile:
            columns.append(str(prof.expr))
            if error:
                columns.append("err({0})".format(str(prof.expr)))

        if normalized:
            binwidths = numpy.full(self._shape[:-1], numpy.inf)

            groupbins = 1.0
            for axis in self._group:
                if isinstance(axis, histbook.axis.groupbin):
                    groupbins *= axis.binwidth
            binwidths[tuple(axis.finiteslice for axis in self._group + self._fixed)] = groupbins

            for i, axis in enumerate(self._fixed):
                if callable(axis.binwidth):
                    binwidth = numpy.array([axis.binwidth(k) for k in range(axis.numbins)])
                else:
                    binwidth = axis.binwidth

                binwidths[tuple(axis.finiteslice if i == j else slice(None) for j, axis in enumerate(self._fixed))] *= binwidth

        def handlearray(content):
            content = content.reshape((-1, self._shape[-1]))

            out = numpy.zeros((content.shape[0], len(columns)), dtype=content.dtype)
            outindex = 0
            
            sumw = content[:, self._sumwindex]
            if count:
                out[:, outindex] = sumw
                countindex = outindex
                outindex += 1

                if error:
                    if self._weightparsed is None:
                        out[:, outindex] = numpy.sqrt(sumw)
                    else:
                        out[:, outindex] = numpy.sqrt(content[:, self._sumw2index])
                    errorindex = outindex
                    outindex += 1

                if normalized:
                    shaped = binwidths.reshape(content.shape[:-1])
                    total = (out[:, countindex] / shaped).sum()
                    correction = total * shaped
                    out[:, countindex] /= correction
                    if error:
                        out[:, errorindex] /= correction

            if len(profile) > 0 or effcount:
                good = sumw > 0
                sumw = sumw[good]
                if self._weightparsed is None:
                    effcnt = sumw
                else:
                    effcnt = numpy.square(sumw) / content[good, self._sumw2index]

            if effcount:
                out[good, outindex] = effcnt
                outindex += 1

            for prof in profile:
                out[good, outindex] = content[good, prof._sumwxindex] / sumw
                outindex += 1
                if error:
                    out[good, outindex] = numpy.sqrt(((content[good, prof._sumwx2index] / sumw) - numpy.square(out[good, outindex - 1])) / effcnt)
                    outindex += 1

            if recarray:
                return out.view([(x, content.dtype) for x in columns]).reshape(self._shape[:-1])
            else:
                return out.reshape(self._shape[:-1] + (outindex,))

        def handle(content):
            if isinstance(content, dict):
                return type(content)((n, handle(x)) for n, x in content.items())
            else:
                return handlearray(content)

        out = handle(self._content)

        if showcolumns:
            return out, columns
        else:
            return out

    def fraction(self, *cut, **opts):
        """
        Return a table of the fraction of entries that pass a set of cuts in each bin.

        Parameters
        ----------
        *cut : :py:class:`profile <histbook.axis.cut>`
            the cut axis or axes to include in the table

        Keyword Arguments
        -----------------
        count : bool
            if ``True`` *(default)*, include the (possibly weighted) count of entries in each bin (denominator of the fraction)

        error : string or ``None``
            if not ``None``, include "errors" on all parameters (uncertainty in the mean of the distribution the count or fraction represents); options are ``"clopper-pearson"``, ``"normal"`` (default), ``"wilson"``, ``"agresti-coull"``, ``"feldman-cousins"``, ``"jeffrey"``, ``"bayesian-uniform"``

        level : number or iterable of numbers
            confidence level or levels at which to evaluate error; default is erf(sqrt(0.5)) or 0.6827, otherwise known as "one sigma"

        recarray : bool
            if ``True`` *(default)*, return results as a Numpy record array, which is rank-2 with named columns; if ``False``, return a plain Numpy array, which is rank-N for N axes and has no column labels.

        columns : bool
            if ``True`` *(not default)*, return a 2-tuple in which the second argument is a list of column labels.
        """
        return self._fraction(cut, opts, False)

    ### FIXME: express levels in sigmas, rather than confidence levels

    def _fraction(self, cut, opts, return_denomhist):
        count = opts.pop("count", True)
        error = opts.pop("error", "normal")
        levels = opts.pop("level", math.erf(math.sqrt(0.5)))
        if isinstance(levels, (numbers.Real, numpy.floating, numpy.integer)):
            levels = (levels,)
        recarray = opts.pop("recarray", True)
        showcolumns = opts.pop("columns", False)
        if len(opts) > 0:
            raise TypeError("unrecognized options for Hist.table: {0}".format(" ".join(opts)))

        self._prefill()

        cut = [x if isinstance(x, histbook.axis.Axis) else self.axis.cut(x) for x in cut]

        cutindex = []
        for c in cut:
            if c not in self._fixed or not isinstance(c, histbook.axis.cut):
                raise IndexError("no such cut axis: {0}".format(c))
            else:
                cutindex.append(self._fixed.index(c))
        cut = [self._fixed[i] for i in cutindex]

        noprofiles = self.drop(*self._profile)
        denomhist = noprofiles.project(*[x for x in self._group + self._fixed if x not in cut])
        cuthist = []
        for c in cut:
            proj = noprofiles.project(*[x for x in self._group + self._fixed if x is c or x not in cut])
            for cutaxis in proj._fixed:
                if isinstance(cutaxis, histbook.axis.cut) and c._parsed == cutaxis._parsed:
                    break
            else:
                raise AssertionError(proj._fixed)
            cuthist.append(proj._selectaxis(cutaxis, histbook.axis._nullaxis(), slice(1, 2), False))

        columns = []
        if count:
            columns.append("count()")
        for c in cut:
            columns.append(str(c.expr))
            if error:
                for level in levels:
                    if level == math.erf(math.sqrt(0.5)):
                        columns.append("err({0})".format(str(c.expr)))
                    else:
                        columns.append("err({0}, {1:3g})".format(str(c.expr), level))

        def level2sigmas(level):
            try:
                from scipy.special import erfinv
            except ImportError:
                a3 = math.pi / 12.0
                a5 = math.pi**2 * 7.0/480.0
                a7 = math.pi**3 * 127.0/40320.0
                a9 = math.pi**4 * 4369.0/5806080.0
                a11 = math.pi**5 * 34807.0/182476800.0
                return 0.5*math.sqrt(math.pi)*(level + a3*level**3 + a5*level**5 + a7*level**7 + a9*level**9 + a11*level**11)
            else:
                return float(erfinv(level) * math.sqrt(2))

        def handlearray(denomcontent, cutcontent):
            denomcontent = denomcontent.reshape((-1, denomhist._shape[-1]))

            out = numpy.zeros((denomcontent.shape[0], len(columns)), dtype=numpy.float64)
            outindex = 0

            sumw = denomcontent[:, denomhist._sumwindex]
            if count:
                out[:, outindex] = sumw
                outindex += 1

            good = sumw > 0
            denom = sumw[good]
            # if denomhist._weightoriginal is not None:
            #     denomw2 = denomcontent[good, denomhist._sumw2index]

            for i in range(len(cut)):
                cc = cutcontent[i].reshape((-1, cuthist[i]._shape[-1]))
                p = out[good, outindex] = cc[good, cuthist[i]._sumwindex] / denom
                outindex += 1

                if not error:
                    pass

                # https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval
                # https://root.cern.ch/root/html602/src/TEfficiency.cxx.html

                elif error == "clopper-pearson":    # recommended by PDG; FIXME: should be the default
                    raise NotImplementedError

                elif error == "normal":
                    for level in levels:
                        out[good, outindex] = level2sigmas(level) * numpy.sqrt(p*(1.0 - p) / denom)
                        outindex += 1

                elif error == "wilson":
                    for level in levels:
                        z = level2sigmas(level)
                        out[good, outindex] = (p + 0.5*z**2/denom + z*numpy.sqrt(p*(1.0 - p)/denom + 0.25*z**2/numpy.square(denom))) / (1.0 + z**2/denom)
                        outindex += 1

                elif error == "agresti-coull":
                    raise NotImplementedError

                elif error == "feldman-cousins":
                    raise NotImplementedError

                elif error == "jeffrey":
                    raise NotImplementedError

                elif error == "bayesian-uniform":
                    raise NotImplementedError

            if recarray:
                return out.view([(x, numpy.dtype(numpy.float64)) for x in columns]).reshape(denomhist._shape[:-1])
            else:
                return out.reshape(denomhist._shape[:-1] + (outindex,))

        def handle(denomcontent, cutcontent):
            if isinstance(denomcontent, dict):
                return type(denomcontent)((n, handle(x, [cc[n] for cc in cutcontent])) for n, x in denomcontent.items())
            else:
                return handlearray(denomcontent, cutcontent)

        out = handle(denomhist._content, [x._content for x in cuthist])

        if showcolumns:
            out = out, columns

        if return_denomhist:
            return out, denomhist
        else:
            return out



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import numpy

from histbook.axis import *
from histbook.hist import *
from histbook.book import *

class TestBook(unittest.TestCase):
    def runTest(self):
        pass

    def test_fill(self):
        b = Book()
        b["one"] = Hist(bin("x", 2, 0, 3, underflow=False, overflow=False, nanflow=False))
        b["two"] = Hist(split("x", 1.5, nanflow=False))
        b.fill(x=[1, 1, 1, 2, 2])
        self.assertEqual(b["one"]._content.tolist(), [[3], [2]])
        self.assertEqual(b["two"]._content.tolist(), [[3], [2]])

        b = Book()
        b["one"] = Hist(bin("x", 2, 0, 3, underflow=False, overflow=False, nanflow=False))
        b["two"] = Hist(split("y", 1.5, nanflow=False))
        b.fill(x=[1, 1, 1, 2, 2], y=[1, 1, 1, 2, 2])
        self.assertEqual(b["one"]._content.tolist(), [[3], [2]])
        self.assertEqual(b["two"]._content.tolist(), [[3], [2]])

    def test_hierarchy(self):
        h = Hist(bin("x", 100, -5, 5))
        outer = Book()
        outer["one"] = h
        outer["two/three"] = h
        self.assertEqual(outer["one"], h)
        self.assertEqual(outer["two/three"], h)
        self.assertEqual(outer["two"]["three"], h)
        self.assertEqual(len(outer["two"]), 1)
        del outer["two/three"]
        self.assertEqual(len(outer["two"]), 0)

    def test_match(self):
        h = Hist(bin("x", 100, -5, 5))
        outer = Book()
        outer["one-a"] = h
        outer["one-b"] = h
        outer["one-c"] = h
        outer["two/three-a"] = h
        outer["two/three-b"] = h
        self.assertEqual(len(outer["one*"]), 3)
        self.assertEqual(len(outer["tw*/*"]), 2)

    def test_add(self):
        book1, book2 = Book(), Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])
        book2["a"] = Hist(bin("x", 4, 0, 4), fill=[2, 2])
        book2["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 3, 3, 3, 3])
        book2["c"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 2, 3])

        book = book1 + book2
        self.assertEqual(book["a"][:].tolist(), [[0], [3], [0], [2], [0], [0], [0]])
        self.assertEqual(book["b"][:].tolist(), [[0], [2], [2], [0], [4], [0], [0]])
        self.assertEqual(book["d"][:].tolist(), [[0], [0], [0], [0], [0], [0], [0]])
        self.assertEqual(book["c"][:].tolist(), [[0], [1], [1], [1], [1], [0], [0]])

    def test_iadd(self):
        book1, book2 = Book(), Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])
        book2["a"] = Hist(bin("x", 4, 0, 4), fill=[2, 2])
        book2["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 3, 3, 3, 3])
        book2["c"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 2, 3])

        book1 += book2
        self.assertEqual(book1["a"][:].tolist(), [[0], [3], [0], [2], [0], [0], [0]])
        self.assertEqual(book1["b"][:].tolist(), [[0], [2], [2], [0], [4], [0], [0]])
        self.assertEqual(book1["d"][:].tolist(), [[0], [0], [0], [0], [0], [0], [0]])
        self.assertEqual(book1["c"][:].tolist(), [[0], [1], [1], [1], [1], [0], [0]])

    def test_mul(self):
        book1 = Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])

        book = book1 * 1.5
        self.assertEqual(book["a"][:].tolist(), [[0], [4.5], [0], [0], [0], [0], [0]])
        self.assertEqual(book["b"][:].tolist(), [[0], [1.5], [3], [0], [0], [0], [0]])
        self.assertEqual(book["d"][:].tolist(), [[0], [0], [0], [0], [0], [0], [0]])

    def test_rmul(self):
        book1 = Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])

        book = 1.5 * book1
        self.assertEqual(book["a"][:].tolist(), [[0], [4.5], [0], [0], [0], [0], [0]])
        self.assertEqual(book["b"][:].tolist(), [[0], [1.5], [3], [0], [0], [0], [0]])
        self.assertEqual(book["d"][:].tolist(), [[0], [0], [0], [0], [0], [0], [0]])

    def test_imul(self):
        book1 = Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])

        book1 *= 1.5
        self.assertEqual(book1["a"][:].tolist(), [[0], [4.5], [0], [0], [0], [0], [0]])
        self.assertEqual(book1["b"][:].tolist(), [[0], [1.5], [3], [0], [0], [0], [0]])
        self.assertEqual(book1["d"][:].tolist(), [[0], [0], [0], [0], [0], [0], [0]])

    def test_group(self):
        book1, book2 = Book(), Book()
        book1["a"] = Hist(bin("x", 4, 0, 4), fill=[0, 0, 0])
        book1["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 1])
        book1["d"] = Hist(bin("x", 4, 0, 4), fill=[])
        book2["a"] = Hist(bin("x", 4, 0, 4), fill=[2, 2])
        book2["b"] = Hist(bin("x", 4, 0, 4), fill=[0, 3, 3, 3, 3])
        book2["c"] = Hist(bin("x", 4, 0, 4), fill=[0, 1, 2, 3])

        book = Book.group(x=book1, y=book2)

        self.assertEqual(book["a"].groupkeys("source"), set(["x", "y"]))
        self.assertEqual(book["b"].groupkeys("source"), set(["x", "y"]))
        self.assertEqual(book["c"].groupkeys("source"), set(["y"]))
        self.assertEqual(book["d"].groupkeys("source"), set(["x"]))

    def test_views(self):
        everything = ChannelsBook(
            mass = SamplesBook(["data", "signal", "background"],
                               SystematicsBook(Hist(bin("x", 5, 0, 5), systematic=[0]),
                                               Hist(bin("x + epsilon", 5, 0, 5), systematic=[1]),
                                               Hist(bin("x - epsilon", 5, 0, 5), systematic=[-1]))),
            truth = SamplesBook(["signal", "background"],
                                Book(par1=Hist(bin("par1", 5, 0, 5)),
                                     par2=Hist(bin("par2", 5, 0, 5)))))

        everything.view("*/data/*").fill(x=numpy.random.uniform(0, 5, 100000), epsilon=numpy.random.normal(0, 0.01, 100000))

        self.assertNotEqual(everything["mass/data/0/0"].table(recarray=False).sum(), 0)
        self.assertEqual(everything["mass/signal/0/0"].table(recarray=False).sum(), 0)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import numbers
import threading

import numpy

import histbook.axis
import histbook.calc
import histbook.calc.spark
import histbook.export
import histbook.expr
import histbook.fill
import histbook.proj
import histbook.instr
import histbook.util
import histbook.vega

class Hist(histbook.fill.Fillable, histbook.proj.Projectable, histbook.export.Exportable, histbook.vega.PlottingChain):
    COUNTTYPE = numpy.float64

    @property
    def _source(self):
        return self

    @property
    def _chain(self):
        return ()

    def weight(self, expr):
        """Returns a copy of this histogram with ``expr`` as weights (for fluent construction)."""
        return Hist(*(self._group + self._fixed + self._profile), weight=expr, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))

    def filter(self, expr):
        """Returns a copy of this histogram with ``expr`` as filter (for fluent construction)."""
        return Hist(*(self._group + self._fixed + self._profile), weight=self._weightoriginal, filter=expr, defs=dict(self._defs), attachment=dict(self._attachment))

    def systematic(self, vector):
        """Returns a copy of this histogram with ``vector`` as systematic (for fluent construction)."""
        return Hist(*(self._group + self._fixed + self._profile), weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment), systematic=vector)

    @classmethod
    def _copycontent(cls, content):
        if content is None:
            return None
        elif isinstance(content, numpy.ndarray):
            return content.copy()
        else:
            return dict((n, cls._copycontent(x)) for n, x in content.items())

    def copy(self):
        """Return an immediate copy of the histogram."""
        out = Hist(*(self._group + self._fixed + self._profile), weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        out._content = self.__class__._copycontent(self._content)
        return out

    def copyonfill(self):
        """Return a copy of the histogram whose content is copied if filled."""
        out = Hist(*(self._group + self._fixed + self._profile), weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))
        out._copyonfill = True
        out._content = self._content
        return out

    def clear(self):
        """Effectively reset all bins to zero."""
        self._content = None

    def cleared(self):
        """Return a copy with all bins set to zero."""
        return Hist(*(self._group + self._fixed + self._profile), weight=self._weightoriginal, filter=self._filteroriginal, defs=dict(self._defs), attachment=dict(self._attachment))        

    def __init__(self, *axis, **opts):
        u"""
        Parameters
        ----------
        *axis : :py:class:`Axis <histbook.axis.Axis>`
            axis or axes that define the independent and dependent variables of the histogram

        Keyword Arguments
        -----------------
        weight : ``None``, algebraic expression (string) or number
            if ``None`` *(default)*, data will be filled with weight ``1``; if an expression, weights are computed from the expression; if a number, weights are constant

        filter : ``None``, algebraic expression (string)
            if not ``None``, data will be filtered through this logical expression (equivalent to multiplying ``weight`` by ``where(filter, 1, 0)``)

        defs : ``None`` or dict of str \u2192 algebraic expression (string) or :py:class:`Expr <histbook.expr.Expr>`
            if not ``None``, definitions to use when computing expressions

        fill : ``None``, single Numpy array or dict of str \u2192 Numpy arrays
            if not ``None``, data to immediately fill after constructing the histogram; single Numpy array is only permitted if there's only one field
            
        attachment : ``None`` or dict of str \u2192 any JSON
            histogram metadata, such as fit results or other context

        systematic : ``None``, tuple of numbers
            the systematic error vector this histogram represents; a special case of attachment (and stored in attachment)
        """
        weight = opts.pop("weight", None)
        filter = opts.pop("filter", None)
        defs = opts.pop("defs", None)
        fill = opts.pop("fill", None)
        attachment = opts.pop("attachment", None)
        systematic = opts.pop("systematic", None)
        if len(opts) > 0:
            raise TypeError("unrecognized options for Hist: {0}".format(" ".join(opts)))

        if systematic is not None:
            if attachment is None:
                attachment = {"systematic": systematic}
            elif "systematic" in attachment:
                raise ValueError("systematic provided as a keyword option and also found in attachment")
            else:
                attachment["systematic"] = systematic

        if defs is None:
            self._defs = {}
        else:
            self._defs = defs
        if attachment is None:
            self._attachment = {}
        else:
            self._attachment = attachment
        self._group = []
        self._fixed = []
        self._profile = []

        newaxis = []
        for old in axis:
            if isinstance(old, histbook.axis._nullaxis) or hasattr(old, "_parsed"):
                newaxis.append(old)
            else:
                expr = histbook.expr.Expr.parse(old._expr, defs=defs)
                new = old.copy()
                new._parsed = expr
                newaxis.append(new)

        self._goals = set()
        self._destination = [[]]
        self._lookup = {}
        def dest(goals):
            self._goals.update(set(goals))
            for goal in goals:
                if goal.goal not in self._lookup:
                    self._lookup[goal.goal] = []
                self._lookup[goal.goal].append(len(self._destination[0]))
                self._destination[0].append(None)

        dictindex = 0
        for new in newaxis:
            if isinstance(new, histbook.axis.GroupAxis):
                self._group.append(new)
                new._dictindex = dictindex
                dictindex += 1
                dest(new._goals(new._parsed))

        self._shape = []
        for new in newaxis:
            if isinstance(new, histbook.axis.FixedAxis):
                self._fixed.append(new)
                new._shapeindex = len(self._shape)
                self._shape.append(new.totbins)
                if not isinstance(new, histbook.axis._nullaxis):
                    dest(new._goals(new._parsed))

        self._shape.append(0)
        for new in newaxis:
            if isinstance(new, histbook.axis.ProfileAxis):
                self._profile.append(new)
                new._sumwxindex = self._shape[-1]
                new._sumwx2index = self._shape[-1] + 1
                self._shape[-1] += 2
                dest(new._goals(new._parsed))

        if filter is None:
            self._filteroriginal, self._filterparsed = None, None

        else:
            self._filteroriginal = filter
            self._filterparsed = histbook.expr.Expr.parse(filter, defs=self._defs)

            if not isinstance(self._filterparsed, (histbook.expr.LogicalOr, histbook.expr.LogicalAnd, histbook.expr.Relation, histbook.expr.Predicate)) and not (isinstance(self._filterparsed, histbook.expr.Const) and self._filterparsed.value in (False, True)):
                raise TypeError("filter must be a logical (boolean) expression") 

        if weight is None and self._filteroriginal is None:
            self._weightoriginal, self._weightparsed = None, None
            self._sumwindex = self._shape[-1]
            self._shape[-1] += 1

        elif weight is None:
            self._weightoriginal = None
            self._weightparsed = histbook.expr.Call("where", self._filterparsed, histbook.expr.Const(1), histbook.expr.Const(0))
            self._sumwindex = self._shape[-1]
            self._sumw2index = self._shape[-1] + 1
            self._shape[-1] += 2
            dest([histbook.instr.CallGraphGoal(self._weightparsed),
                  histbook.instr.CallGraphGoal(histbook.expr.Call("numpy.multiply", self._weightparsed, self._weightparsed))])

        elif self._filteroriginal is None:
            self._weightoriginal = weight
            if isinstance(weight, (numbers.Real, numpy.integer, numpy.floating)):
                self._weightparsed = histbook.expr.Const(weight)
            else:
                self._weightparsed = histbook.expr.Expr.parse(weight, defs=self._defs)
            self._sumwindex = self._shape[-1]
            self._sumw2index = self._shape[-1] + 1
            self._shape[-1] += 2
            dest([histbook.instr.CallGraphGoal(self._weightparsed),
                  histbook.instr.CallGraphGoal(histbook.expr.Call("numpy.multiply", self._weightparsed, self._weightparsed))])

        else:
            self._weightoriginal = weight
            if isinstance(weight, (numbers.Real, numpy.integer, numpy.floating)):
                weight = repr(weight)
            self._weightparsed = histbook.expr.Expr.parse("where(({0}), 1, 0) * ({1})".format(self._filteroriginal, weight), defs=self._defs)
            self._sumwindex = self._shape[-1]
            self._sumw2index = self._shape[-1] + 1
            self._shape[-1] += 2
            dest([histbook.instr.CallGraphGoal(self._weightparsed),
                  histbook.instr.CallGraphGoal(histbook.expr.Call("numpy.multiply", self._weightparsed, self._weightparsed))])

        self._group = tuple(self._group)
        self._fixed = tuple(self._fixed)
        self._profile = tuple(self._profile)

        self._shape = tuple(self._shape)
        self._content = None
        self._fields = None
        self._copyonfill = False

        if fill is not None:
            if not histbook.calc.spark.isspark(fill, {}) and not isinstance(fill, dict):
                if len(self._group + self._fixed + self._profile) == 1:
                    fill = {str((self._group + self._fixed + self._profile)[0]._parsed): fill}
                else:
                    raise TypeError("fill must be a dict for histograms of more than one axis")
            self.fill(fill)

    @property
    def defs(self):
        """Definitions used by axis expressions."""
        return self._defs

    def attach(self, key, value):
        """Add an attachment to the histogram (changing it in-place and returning it)."""
        self._attachment[key] = value
        return self

    def detach(self, key):
        """Remove an attachment from the histogram (changing it in-place and returning it)."""
        del self._attachment[key]
        return self

    def has(self, key):
        """Returns ``True`` if ``key`` exists in the attachment metadata."""
        return key in self._attachment

    def get(self, key, *default):
        """
        Get an item of attachment metadata.

        If ``key`` isn't found and no ``default`` is specified, raise a ``KeyError``.
        If ``key`` isn't found and a ``default`` is provided, return the ``default`` instead.

        Only one ``default`` is allowed.
        """
        if len(default) == 0:
            return self._attachment[key]
        elif len(default) == 1:
            return self._attachment.get(key, default[0])
        else:
            raise TypeError("get takes 1 or 2 arguments; {0} provided".format(len(default) + 1))

    @property
    def attachment(self):
        """Python dict of attachment metadata (linked, not a copy)."""
        return self._attachment

    def __repr__(self, indent=", "):
        out = [repr(x) for x in self._group + self._fixed + self._profile]
        if self._weightoriginal is not None:
            out.append("weight={0}".format(repr(self._weightoriginal)))
        if self._filteroriginal is not None:
            out.append("filter={0}".format(repr(self._filteroriginal)))
        if len(self._defs) > 0:
            out.append("defs={" + ", ".join("{0}: {1}".format(repr(n), repr(str(x)) if isinstance(x, histbook.expr.Expr) else repr(x)) for n, x in self._defs.items()) + "}")
        return "Hist(" + indent.join(out) + ")"

    def __str__(self, indent=",\n     ", first=""):
        return self.__repr__(indent)

    @property
    def shape(self):
        """Shape of the Numpy array defining the content of the fixed-memory axes (:py:class:`FixedAxis <histbook.axis.FixedAxis>`) only."""
        return self._shape

    def _streamline(self, i, instructions):
        for instruction in instructions:
            if isinstance(instruction, histbook.instr.Export):
                if not hasattr(instruction, "destination"):
                    instruction.destination = []
                if instruction.goal in self._lookup:
                    for j in self._lookup[instruction.goal]:
                        instruction.destination.append((i, j))

        return instructions

    def fill(self, arrays=None, **more):
        u"""
        Fill the histogram: identify bins for independent variables, increase their counts by ``1`` or ``weight``, and increment any profile (dependent variable) means and errors in the means.

        All arrays must have the same length (one-dimensional shape). Numbers are treated as one-element arrays.

        Parameters
        ----------
        arrays : dict \u2192 Numpy array or number; Spark DataFrame; Pandas DataFrame
            field values to use in the calculation of independent and dependent variables (axes)

        **more : Numpy arrays or numbers
            more field values
        """
        if self._copyonfill:
            self._content = Hist._copycontent(self._content)
            self._copyonfill = False

        if histbook.calc.spark.isspark(arrays, more):
            # pyspark.DataFrame
            wait = histbook.calc.spark.fillspark(self, arrays)
            self._prefill()
            wait()

        elif arrays.__class__.__name__ == "DataFrame" and arrays.__class__.__module__ == "pandas.core.frame":
            # pandas.DataFrame
            if len(more) > 0:
                raise TypeError("if arrays is a Pandas DataFrame, keyword arguments are not allowed")
            self.fill(dict((n, arrays[n].values) for n in arrays.columns))

        else:
            # dict-like of numpy.ndarray (or castable)
            if arrays is None:
                arrays = more
            elif len(more) == 0:
                pass
            else:
                arrays = histbook.util.ChainedDict(arrays, more)

            self._prefill()

            length = self._fill(arrays)
            self._postfill(arrays, length)

    def _prefill(self):
        if self._content is None:
            if len(self._group) == 0:
                self._content = numpy.zeros(self._shape, dtype=self.COUNTTYPE)

            elif isinstance(self._group[0], histbook.axis.groupby) and self._group[0].keeporder:
                self._content = collections.OrderedDict()

            else:
                self._content = {}

    def _postfill(self, arrays, length):
        j = len(self._group)
        step = 0
        indexes = None
        for axis in self._fixed:
            if step == 0:
                indexes = self._destination[0][j]
            elif step == 1:
                indexes = indexes.copy()
            if step > 0:
                numpy.multiply(indexes, self._shape[axis._shapeindex], indexes)
                numpy.add(indexes, self._destination[0][j], indexes)
            j += 1
            step += 1

        axissumx, axissumx2 = [], []
        for axis in self._profile:
            axissumx.append(self._destination[0][j])
            axissumx2.append(self._destination[0][j + 1])
            j += 2

        if self._weightparsed is None:
            weight = 1
            weight2 = None
        elif isinstance(self._weightparsed, histbook.expr.Const):
            weight = numpy.ones(length) * self._weightparsed.value
            weight2 = numpy.ones(length) * self._weightparsed.value**2
        else:
            weight = self._destination[0][j]
            weight2 = self._destination[0][j + 1]
            selection = numpy.isnan(weight)
            if selection.any():
                weight = weight.copy()
                weight2 = weight2.copy()
                weight[selection] = 0.0
                weight2[selection] = 0.0

        def fillblock(content, indexes, axissumx, axissumx2, weight, weight2):
            for sumx, sumx2, axis in zip(axissumx, axissumx2, self._profile):
                if indexes is None:
                    indexes = numpy.ma.zeros(len(sumx), dtype=histbook.calc.INDEXTYPE)
                if weight2 is not None:
                    selection = numpy.ma.getmask(indexes)
                    if selection is not numpy.ma.nomask:
                        selection = numpy.bitwise_not(selection)
                        weight = weight[selection]
                        weight2 = weight2[selection]
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, axis._sumwxindex], indexes.compressed(), sumx * weight)
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, axis._sumwx2index], indexes.compressed(), sumx2 * weight)

            if weight2 is None:
                if indexes is None:
                    content.reshape((-1, self._shape[-1]))[:, self._sumwindex] += (1 if length is None else length) * weight
                else:
                    numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumwindex], indexes.compressed(), weight)
            else:
                if indexes is None:
                    indexes = numpy.ma.zeros(len(weight), dtype=histbook.calc.INDEXTYPE)
                selection = numpy.ma.getmask(indexes)
                if selection is not numpy.ma.nomask:
                    selection = numpy.bitwise_not(selection)
                    weight = weight[selection]
                    weight2 = weight2[selection]
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumwindex], indexes.compressed(), weight)
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumw2index], indexes.compressed(), weight2)

        def filldict(j, content, indexes, axissumx, axissumx2, weight, weight2, allselection):
            if j == len(self._group):
                fillblock(content, indexes, axissumx, axissumx2, weight, weight2)

            else:
                uniques, inverse = self._destination[0][j]
                for idx, unique in enumerate(uniques):
                    if allselection is None:
                        selection = (inverse == idx)
                    else:
                        selection = (inverse[allselection] == idx)
                    
                    if unique not in content:
                        if j + 1 == len(self._group):
                            content[unique] = numpy.zeros(self._shape, dtype=self.COUNTTYPE)

                        elif isinstance(self._group[j + 1], histbook.axis.groupby) and self._group[j + 1].keeporder:
                            content[unique] = collections.OrderedDict()

                        else:
                            content[unique] = {}

                    subcontent = content[unique]
                    if indexes is None:
                        subindexes = numpy.ma.zeros(numpy.count_nonzero(selection), dtype=histbook.calc.INDEXTYPE)
                    else:
                        subindexes = indexes[selection]
                    subaxissumx = [x[selection] for x in axissumx]
                    subaxissumx2 = [x[selection] for x in axissumx2]
                    if weight2 is None:
                        subweight, subweight2 = weight, weight2
                    else:
                        subweight = weight[selection]
                        subweight2 = weight2[selection]

                    if allselection is None:
                        suballselection = selection
                    else:
                        suballselection = allselection.copy()
                        suballselection[inverse != idx] = False

                    filldict(j + 1, subcontent, subindexes, subaxissumx, subaxissumx2, subweight, subweight2, suballselection)

        filldict(0, self._content, indexes, axissumx, axissumx2, weight, weight2, None)

        for j in range(len(self._destination[0])):
            self._destination[0][j] = None

    def __add__(self, other):
        if not isinstance(other, Hist):
            raise TypeError("histograms can only be added to other histograms")

        if self._group + self._fixed + self._profile != other._group + other._fixed + other._profile:
            raise TypeError("histograms can only be added to other histograms with the same axis specifications")

        def add(selfcontent, othercontent):
            if selfcontent is None and othercontent is None:
                return None

            elif selfcontent is None:
                return Hist._copycontent(othercontent)

            elif othercontent is None:
                return Hist._copycontent(selfcontent)

            elif isinstance(selfcontent, numpy.ndarray) and isinstance(othercontent, numpy.ndarray):
                return selfcontent + othercontent

            else:
                assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
                out = {}
                for n in selfcontent:
                    if n in othercontent:
                        out[n] = add(selfcontent[n], othercontent[n])
                    else:
                        out[n] = selfcontent[n]
                for n in othercontent:
                    if n not in selfcontent:
                        out[n] = othercontent[n]
                return out

        out = self.__class__.__new__(self.__class__)
        out.__dict__.update(self.__dict__)
        out._content = add(self._content, other._content)
        return out

    def __iadd__(self, other):
        if not isinstance(other, Hist):
            raise TypeError("histograms can only be added to other histograms")

        if self._group + self._fixed + self._profile != other._group + other._fixed + other._profile:
            raise TypeError("histograms can only be added to other histograms with the same axis specifications")

        def add(selfcontent, othercontent):
            assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
            for n in selfcontent:
                if n in othercontent:
                    if isinstance(selfcontent[n], numpy.ndarray):
                        selfcontent[n] += othercontent[n]
                    else:
                        add(selfcontent[n], othercontent[n])
            for n in othercontent:
                if n not in selfcontent:
                    selfcontent[n] = Hist._copycontent(othercontent[n])

        if other._content is None:
            pass

        elif self._content is None:
            self._content = Hist._copycontent(other._content)

        elif isinstance(self._content, numpy.ndarray):
            self._content += other._content

        else:
            add(self._content, other._content)

        return self

    def __mul__(self, value):
        if not isinstance(value, (numbers.Real, numpy.integer, numpy.floating)):
            raise TypeError("Hist can only be multiplied by a scalar number.")

        def recurse(content):
            if isinstance(content, dict):
                return dict((n, recurse(x)) for n, x in content.items())
            else:
                return content * value

        out = self.__class__.__new__(self.__class__)
        out.__dict__.update(self.__dict__)
        out._content = recurse(self._content)
        return out

    def __rmul__(self, value):
        return self.__mul__(value)

    def __imul__(self, value):
        if not isinstance(value, (numbers.Real, numpy.integer, numpy.floating)):
            raise TypeError("Hist can only be multiplied by a scalar number.")

        def recurse(content):
            if isinstance(content, dict):
                for x in content.values():
                    recurse(x)
            else:
                content *= value

        recurse(self._content)
        return self

    @classmethod
    def group(cls, by="source", **hists):
        """
        Combine histograms, maintaining their distinctiveness by adding a new categorical axis to each.

        To combine histograms by adding bins, just use the ``+`` operator.

        Parameters
        ----------
        by : string
            name of the new axis (must not already exist)

        **hists : :py:class:`Hist <histbook.hist.Hist>`
            histograms to combine (must have the same axes)
        """
        if any(not isinstance(x, Hist) for x in hists.values()):
            raise TypeError("only histograms can be grouped with other histograms")

        axis = None
        hist = None
        for x in hists.values():
            hist = x
            if axis is None:
                axis = x._group + x._fixed + x._profile
            elif axis != x._group + x._fixed + x._profile:
                raise TypeError("histograms can only be grouped with the same axis specifications")
            
        if axis is None:
            raise ValueError("at least one histogram must be provided")

        if histbook.axis.groupby(by) in axis:
            raise ValueError("groupby({0}) already exists in these histograms; use hist.togroup(other) to add to a group".format(repr(by)))

        weight = None
        for x in hists.values():
            if weight is None:
                weight = x._weightoriginal
            elif x._weightoriginal is None:
                raise TypeError("histograms can only be grouped with the same weight specifications")
            elif weight != x._weightoriginal:
                weight = 1.0
                
        defs = {}
        for x in hists.values():
            defs.update(x._defs)

        out = cls(*((histbook.axis.groupby(by),) + hist._group + hist._fixed + hist._profile), weight=weight, filter=None, defs=dict(defs), attachment=None)
        out._content = {}
        for n, x in hists.items():
            out._content[n] = cls._copycontent(x._content)
        return out

    def togroup(**hists):
        u"""
        Add histograms to the :py:class:`groupby <histbook.axis.groupby>` that is the first axis.

        Histograms created with :py:meth:`Hist.group <histbook.hist.Hist.group>` have a first axis that is a :py:class:`groupby <histbook.axis.groupby>`.

        Keyword Arguments
        -----------------
        **hists : dict of str \u2192 :py:class:`Hist <histbook.hist.Hist>`
            histograms to add to the existing group
        """
        if len(self._group) == 0 or not isinstance(self._group[0], histbook.axis.groupby):
            raise ValueError("togroup can only be used on histograms whose first axis is a groupby")

        if any(not isinstance(x, Hist) for x in hists.values()):
            raise TypeError("only histograms can be grouped")

        for x in hists.values():
            if self._group[1:] + self._fixed + self._profile != x._group + x._fixed + x._profile:
                raise TypeError("histograms can only be grouped with the same axis specifications")

        def add(selfcontent, othercontent):
            assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
            for n in selfcontent:
                if n in othercontent:
                    if isinstance(selfcontent[n], numpy.ndarray):
                        selfcontent[n] += othercontent[n]
                    else:
                        add(selfcontent[n], othercontent[n])
            for n in othercontent:
                if n not in selfcontent:
                    selfcontent[n] = Hist._copycontent(othercontent[n])

        for n, x in hists.items():
            if n in self._content:
                add(self._content[n], other._content)
            else:
                self._content[n] = Hist._copycontent(other._content)

    def tojson(self):
        out = {"type": "Hist", "axis": [x.tojson() for x in self._group + self._fixed + self._profile]}
        if self._weightoriginal is not None:
            out["weight"] = self._weightoriginal
        if self._filteroriginal is not None:
            out["filter"] = self._filteroriginal
        if self._defs is not None and len(self._defs) != 0:
            out["defs"] = self._defs
        if self._content is not None:
            def recurse(node):
                if isinstance(node, dict):
                    return dict((n, recurse(x)) for n, x in node.items())
                else:
                    return node.tolist()
            out["content"] = recurse(self._content)
        if len(self._attachment) != 0:
            out["attachment"] = self._attachment
        return out

    @staticmethod
    def fromjson(obj):
        assert obj["type"] == "Hist"
        def recurse(node):
            if node is None:
                return None
            elif isinstance(node, dict):
                return dict((n, recurse(x)) for n, x in node.items())
            else:
                return numpy.array(node, dtype=Hist.COUNTTYPE)

        out = Hist(*[histbook.axis.Axis.fromjson(x) for x in obj["axis"]], weight=obj.get("weight", None), filter=obj.get("filter", None), defs=obj.get("defs", None), attachment=obj.get("attachment", None))
        out._content = recurse(obj.get("content", None))
        return out

    def __getstate__(self):
        packed = tuple(x._pack() for x in self._group + self._fixed + self._profile)
        return (packed, self._weightoriginal, self._filteroriginal, None if len(self._defs) == 0 else self._defs, self._content, None if len(self._attachment) == 0 else self._attachment)

    def __setstate__(self, state):
        packed, weight, filter, defs, content, attachment = state
        self.__init__(*[histbook.axis.Axis._unpack(x) for x in packed], weight=weight, filter=filter, defs=defs, attachment=attachment)
        self._content = content

    def __eq__(self, other):
        def recurse(one, two):
            if one is None and two is None:
                return True
            elif isinstance(one, dict) and isinstance(two, dict):
                return set(one.keys()) == set(two.keys()) and all(recurse(one[n], two[n]) for n in one)
            elif isinstance(one, numpy.ndarray) and isinstance(two, numpy.ndarray):
                return numpy.array_equal(one, two)
            else:
                return False

        return self.__class__ is other.__class__ and self._group == other._group and self._fixed == other._fixed and self._profile == other._profile and self._weightparsed == other._weightparsed and self._filterparsed == other._filterparsed and self._defs == other._defs and recurse(self._content, other._content) and self._attachment == other._attachment

    def __ne__(self, other):
        return not self.__eq__(other)

    def compatible(self, other):
        """Returns True if the histograms have the same non-profile axis types and binning, regardless of the expressions used to compute them."""
        return len(self._group) == len(other._group) and len(self._fixed) == len(other._fixed) and all(y.compatible(z) for y, z in zip(self._group + self._fixed, other._group + other._fixed))

    def __getitem__(self, where):
        if not isinstance(where, tuple):
            where = (where,)
        self._prefill()
        out = self._content
        for i in where:
            out = out[i]
        return out

    # a similar __setitem__ method would require checks to ensure the user doesn't mess up the structure

    def groupkeys(self, axis):
        """
        Return all categorical keys associated with a groupby axis or non-zero bins associated with a groupbin axis.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (string), or index position (integer)
            the groupby or groupbin axis

        Returns
        -------
        set
            all keys for this axis, even if that is a union over other group axes        
        """
        if not isinstance(axis, histbook.axis.Axis):
            axis = self.axis[axis]
        for i, x in enumerate(self._group):
            if x == axis:
                break
        else:
            raise IndexError("no such groupby/groupbin axis: {0}".format(x))

        out = set()
        def recurse(j, content):
            if i == j:
                out.update(content.keys())
            else:
                for x in content.values():
                    recurse(j + 1, x)
        
        recurse(0, self._content)
        return out
